#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Module implementing the PyFuncApp class"""

import ast
import base64
import collections
import importlib
import inspect
import json
import logging
import os
import pickle

from typing import Callable, Union
import dill
from io import StringIO
from contextlib import redirect_stdout

from dlg import drop_loaders
from dlg.data.path_builder import filepath_from_string
from dlg.drop import track_current_drop
from dlg.utils import deserialize_data
from dlg.named_port_utils import (
    Argument,
    DropParser,
    check_ports_dict,
    get_port_reader_function,
    identify_named_ports,
    resolve_drop_parser,
)
from dlg.apps.app_base import BarrierAppDROP
from dlg.exceptions import BadModuleException, IncompleteDROPSpec, InvalidPathException
from dlg.meta import (
    dlg_string_param,
    dlg_dict_param,
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
)
from dlg.pyext import pyext

logger = logging.getLogger(f"dlg.{__name__}")

MAX_IMPORT_RECURSION = 100


def import_using_name(app, fname, curr_depth):
    if curr_depth > MAX_IMPORT_RECURSION:
        raise BadModuleException(
            app, "Problem importing module %s, search exceeded recursion limit" % fname
        )

    logger.debug("Importing %s", fname)
    parts = fname.split(".")
    # If only one part check if builtin
    b = globals()["__builtins__"]
    logger.debug("Function: %s: %s", parts[0], hasattr(b, parts[0]))
    if len(parts) < 2:
        logger.debug("Builtins: %s", type(b))
        logger.debug("Function %s: %s", fname, hasattr(b, fname))
        if fname in b:
            return b[fname]
        else:
            msg = "%s is not builtin and does not contain a module name" % fname
            raise BadModuleException(app, msg)
    elif parts[0] in b.keys():
        return b[parts[0]]
    else:
        if len(parts) > 1:
            if parts[-1] in ["__init__", "__class__"]:
                parts = parts[:-1]
            logger.debug("Recursive import: %s", parts)
            try:
                mod = importlib.import_module(parts[0], __name__)
            except ImportError as e:
                raise BadModuleException(
                    app,
                    "Error when loading module %s: %s" % (parts[0], str(e)),
                ) from e
            for m in parts[1:]:
                try:
                    logger.debug("Getting attribute %s", m)
                    mod = getattr(mod, m)
                except AttributeError:
                    try:
                        logger.debug(
                            "Trying to load backwards: %s",
                            ".".join(parts[:-1]),
                        )
                        mod = importlib.import_module(".".join(parts[:-1]), __name__)
                        mod = getattr(mod, parts[-1])
                        break
                    except ModuleNotFoundError:
                        # try again, sometimes fixes the namespace
                        mod = import_using_name(app, fname, curr_depth=curr_depth + 1)
                        break
                    except Exception as e:  # pylint: disable=broad-exception-caught
                        raise BadModuleException(
                            app, "Problem importing module %s, %s" % (mod, e)
                        ) from e
            logger.debug("Loaded module: %s", mod)
            return mod


def import_using_code_ser(func_code: Union[str, bytes], func_name: str):
    """
    Import the function provided as a serialised code string.
    """
    try:
        func_code = func_code if isinstance(func_code, bytes) else func_code.encode()
        func = dill.loads(base64.b64decode(func_code))
    except Exception as err:
        logger.warning("Unable to deserialize func_code: %s", err)
        raise
    if func_name and func_name.split(".")[-1] != func.__name__:
        raise ValueError(
            f"Function with name '{func.__name__}' instead of '{func_name}' found!"
        )
    return func


def import_using_code(func_code: str, func_name: str, serialized: bool = True):
    """
    Import the function provided as a code string. Plain code as well as serialized code
    is supported. If the func_name does not match the provided func_name the load will fail.
    """
    mod = None
    if not serialized and not isinstance(func_code, bytes):
        logger.debug("Trying to import code from string: %s\n", func_code.strip())
        try:
            mod = pyext.RuntimeModule.from_string("mod", func_name, func_code.strip())
        except Exception:  # pylint: disable=broad-exception-caught
            func = import_using_code_ser(func_code, func_name)
        logger.debug("Imported function: %s", func_name)
        if mod and func_name:
            if hasattr(mod, func_name):
                func = getattr(mod, func_name)
            else:
                logger.warning("Function with name '%s' not found!", func_name)
                raise ValueError(f"Function with name '{func_name}' not found!")
    else:
        func = import_using_code_ser(func_code, func_name)
    logger.debug("Imported function: %s", func_name)
    return func


##
# @brief PythonMemberFunction
# @details A placeholder APP to aid construction of new class member function applications.
# This is mainly useful (and used) when starting a new workflow from scratch.
# @par EAGLE_START
# @param category PythonMemberFunction
# @param tag daliuge
# @param func_name object.__init__/String/ComponentParameter/NoPort/ReadWrite//False/False/Python function name
# @param func_code /String/ComponentParameter/NoPort/ReadWrite//False/False/Python function code, e.g. 'def f(args): return args'. Function name has to be 'f'!
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.pyfunc.PyFuncApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name pyfunc/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param self /Object/ComponentParameter/InputOutput/ReadWrite//False/False/Port exposing the object
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/The allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
#     \~English Mapping from argname to default value. Should match only the last part of the argnames list.
#               Values are interpreted as Python code literals and that means string values need to be quoted.
# @par EAGLE_END
class PyMemberApp(BarrierAppDROP):
    """A placeholder member function that just aids the generation of the palette component"""


##
# @brief PyFuncApp
# @details An application that wraps a python function.
# The inputs of the application are treated as the arguments of the function.
# Conversely, the output of the function is treated as the output of the
# application. If the application has more than one output, the result of
# calling the function is treated as an iterable, with each individual object
# being written to its corresponding output.
# @par EAGLE_START
# @param category PyFuncApp
# @param tag template
# @param func_name /String/ComponentParameter/NoPort/ReadWrite//False/False/Python function name
# @param func_code /String/ComponentParameter/NoPort/ReadWrite//False/False/Python function code, e.g. 'def func_name(args): return args'. Note that func_name above needs to match the defined name here.
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.pyfunc.PyFuncApp/String/ComponentParameter/NoPort/ReadOnly//False/False/Application class
# @param base_name pyfunc/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param input_error_threshold 0/Integer/ComponentParameter/NoPort/ReadWrite//False/False/The allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param n_tries 1/Integer/ComponentParameter/NoPort/ReadWrite//False/False/Specifies the number of times the 'run' method will be executed before finally giving up
#     \~English Mapping from argname to default value. Should match only the last part of the argnames list.
#               Values are interpreted as Python code literals and that means string values need to be quoted.
# @par EAGLE_END
class PyFuncApp(BarrierAppDROP):
    """
    An application that wraps a simple python function.

    The inputs of the application are treated as the arguments of the function.
    Conversely, the output of the function is treated as the output of the
    application. If the application has more than one output, the result of
    calling the function is treated as an iterable, with each individual object
    being written to its corresponding output.

    Users indicate the function to be wrapped via the ``func_name`` parameter.
    In this case func_name needs to specify a function in the standard form

    ``module.function``

    and the module needs to be accessible on the PYTHONPATH of the DALiuGE
    engine. Note that the engine is expanding the standard PYTHONPATH with
    DLG_ROOT/code. That directory is always available, even if the engine is
    running in a docker container.

    Otherwise, users can also *send* over the python code using the ``func_code``
    parameter. The code needs to be base64-encoded and produced with the marshal
    module of the same Python version used to run DALiuGE.

    The positional onlyargs will be used in order of appearance.
    """

    component_meta = dlg_component(
        "PyFuncApp",
        "Py Func App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    func_name = dlg_string_param("func_name", None)
    func_arg_mapping = dlg_dict_param("func_arg_mapping", {})
    func_defaults = dlg_dict_param("func_defaults", {})
    func: Callable
    fdefaults: dict

    def __repr__(self):
        if hasattr(self, "func"):
            return f"{self.__class__.__name__}_{self.func}_{self.oid}"
        else:
            return f"{self.__class__.__name__}"

    def _mixin_func_defaults(self):
        """
        func_defaults can be passed in through an argument, but this is only used for
        dlg_delayed and in combination with passing func_code. For dlg_delayed the function
        and its parameters might be computed on a different host than the where the delayed
        function is called and thus the function needs to be serialized.
        """
        logger.debug("Starting evaluation of func_defaults: %s", self.func_defaults)
        if (
                isinstance(self.func_defaults, dict)
                and len(self.func_defaults) > 0
                and list(self.func_defaults.keys()) == ["kwargs", "args"]
        ):
            for arg in self.func_defaults["args"]:
                self.func_defaults["kwargs"][arg] = arg
                self.func_defaults = self.func_defaults["kwargs"]
        elif (
                isinstance(self.func_defaults, dict)
                and "kwargs" in self.func_defaults
                and isinstance(self.func_defaults["kwargs"], dict)
        ):
            self.func_defaults = self.func_defaults["kwargs"]
        # we came all this way, now assume that any resulting dict is correct
        if not isinstance(self.func_defaults, dict):
            logger.error(
                "Wrong format or type for function defaults for %s: %r, %r",
                self.func.__name__,
                self.func_defaults,
                type(self.func_defaults),
            )
            raise ValueError
        for name, value in self.func_defaults.items():
            # self.func_defaults[name] = deserialize_data(value) if value else value
            self.func_defaults[name] = deserialize_data(value)
        # the fn_defaults are used afterwards, we'll drop the func_defaults
        logger.debug("fn_defaults %s", self.fn_defaults)
        logger.debug("func_defaults %s", self.func_defaults)
        self.fn_defaults.update(self.func_defaults)

    def _init_fn_defaults(self):
        """
        Inititalize self.fn_defaults dictionary from values provided.
        Multiple options exist and some are here for compatibility.
        """
        logger.debug("Starting evaluation of function signature")
        # set defaults
        self.fn_defaults = {}
        self.fn_nposkw = 0
        self.poskw = {}
        self.fn_npos = 0
        self.posonly = {}
        self.kwonly = {}
        self.fn_nkw = 0
        self.varargs = False
        self.varkw = False
        self.fn_ndef = 0
        self.argnames = []
        self.argsig = None
        self.fn_nargs = 0
        try:
            self.argsig = inspect.signature(self.func)
            self.argnames = list(self.argsig.parameters.keys())
            logger.debug("Function signature: %s", self.argsig)
            self.fn_nargs = len(self.argsig.parameters)
            self.populate_arguments_without_signature = False
        except ValueError:
            logger.warning("Unable to uncover signature from function. This is likely a "
                           "C-based Python function. No positional/keyword arguments "
                           "determined.")
            self.argsig = self.func
            self.populate_arguments_without_signature = True

            return False

        for k, p in self.argsig.parameters.items():
            if not isinstance(p, str):
                if p.kind == p.POSITIONAL_OR_KEYWORD:
                    self.fn_nposkw += 1
                    self.poskw.update({k: p})
                elif p.kind == p.POSITIONAL_ONLY:
                    self.fn_npos += 1
                    self.posonly.update({k: p})
                elif p.kind == p.KEYWORD_ONLY:
                    self.fn_nkw += 1
                    self.kwonly.update({k: p})
                elif p.kind == p.VAR_POSITIONAL:
                    self.varargs = True
                elif p.kind == p.VAR_KEYWORD:
                    self.varkw = True
                if p.default != inspect._empty:  # pylint: disable=protected-access
                    self.fn_defaults[k] = p.default
                    # self.arguments_defaults.append(p.default)
                    self.fn_ndef += 1
                else:
                    self.fn_defaults[k] = None
                    self.arguments_defaults.append(None)
                    self.fn_ndef += 1

        logger.debug("Got signature for function %s %s", self.func, self.argsig)
        logger.debug("Got default values for arguments %s", self.arguments_defaults)
        logger.debug("initialized fn_defaults with %s", self.fn_defaults)
        return True

    def _clean_applicationArgs(self):
        """
        remove function definition arguments in applicationArgs
        """
        self.func_def_keywords = [
            "func_code",
            "func_name",
            "func_arg_mapping",
            "input_parser",
            "output_parser",
            "func_defaults",
            "pickle",
        ]

        for kw in self.func_def_keywords:
            # these are the preferred ones now
            if kw in self._applicationArgs and isinstance(
                    self._applicationArgs[kw]["value"],
                    bool or self._applicationArgs[kw]["value"] or
                    self._applicationArgs[kw]["precious"],
            ):
                # only transfer if there is a value or precious is True
                self._applicationArgs.pop(kw)

    def _initialise_args(self, argsDict: dict):
        """
        Ensure default arguments are properly setup.
        # If there are default values in the function sig
        # If there are values in the graph
        """
        for arg, value in argsDict.items():
            argsDict[arg] = Argument(value)
            param_arg = self.parameters.get(arg)
            if param_arg:
                argsDict[arg].value = param_arg
                argsDict[arg].encoding = "dill"
        return argsDict

    def _get_arg_info(self, arg):
        """
        Get necessary information for populating the argument
        """
        value = self._applicationArgs[arg].get("value", None)
        encoding = self._applicationArgs[arg].get("encoding", "dill")
        precious = self._applicationArgs[arg].get("precious", False)
        positional = self._applicationArgs[arg].get("positional", False)
        portType = self._applicationArgs[arg].get("usage", "NoPort")

        return value, encoding, precious, positional, portType

    def _populate_arguments_with_graph_data(self, positionalArgsMap, keywordArgsMap):
        """
         Go through the parameters that were provided by the user to the Logical Graph and
        populate the application argument with the data.

        If the user didn't provide any input, or the data is expected to come through an
        InputPort, the app argument will not be updated.

        :param positionalArgsMap:
        :param keywordArgsMap:

        :return: positionalArgsMap,
        :return: keywordArgsMap
        :return: input_outputs
        """

        vparg = []
        vkarg = {}
        input_outputs = []

        # Port-types that we may receive a filename/path from that is used as an output
        # reference
        filename_ports = ["InputOutput", "OutputPort"]

        # Legacy, used for graphs reliant on input/output parser
        if not self._applicationArgs:
            encoding = DropParser.DILL
            for key in positionalArgsMap:
                positionalArgsMap[key].encoding = encoding
            logger.debug("AppArgs/pargsDict: %s", positionalArgsMap)

            return positionalArgsMap, keywordArgsMap, input_outputs, vparg, vkarg

        # If we overwrite any defaults with what is store in AppArgs "value"
        for arg in self._applicationArgs:
            # check value type and interpret
            value, encoding, precious, positional, portType = self._get_arg_info(
                arg)
            if self._applicationArgs[arg]["type"] in ["Json", "Complex"]:
                try:
                    value = ast.literal_eval(value)
                    logger.debug("Evaluated %s to %s",
                                 value, type(value))
                    self._applicationArgs[arg].value = value
                except ValueError:
                    logger.error("Unable to evaluate %s",
                                 self._applicationArgs[arg]["value"])
            for arg_map in [positionalArgsMap, keywordArgsMap]:
                if arg in arg_map:
                    argument = arg_map[arg]
                    argument.value = value or argument.value
                    argument.encoding = encoding or argument.encoding
                    argument.precious = precious or argument.precious
                    argument.positional = positional or argument.positional
                    arg_map[arg] = argument

            if portType in filename_ports:
                input_outputs.append(arg)

        # Remove parameters of function that have been found in applicationArgs
        _ = [self._applicationArgs.pop(k) for k in positionalArgsMap if
             k in self._applicationArgs]
        logger.debug("Updated posargs dictionary: %s", positionalArgsMap)
        logger.debug("Updated keyargs dictionary: %s", keywordArgsMap)

        # Put all remaining arguments into *args and **kwargs
        logger.debug("Remaining AppArguments %s", self._applicationArgs)
        for arg in self._applicationArgs:
            if self._applicationArgs[arg]["type"] in ["Json", "Complex"]:
                value = ast.literal_eval(self._applicationArgs[arg]["value"])
            else:
                value = self._applicationArgs[arg]["value"]
            if (self._applicationArgs[arg]["positional"] or
                    self.populate_arguments_without_signature):
                vparg.append(value)
            else:
                vkarg[arg] = value

        return positionalArgsMap, keywordArgsMap, input_outputs, vparg, vkarg

    def _map_parameters_to_func_args(self, positionalArgs: dict, keywordArguments: dict) -> tuple:
        """
        Given the keyword and postional arguments of the function we are using, map
        the EAGLE/DALiuGE input parameters to those arguments.

        This performs the following steps:

        :param positionalArgs: postitional arguments of the function defined in self.func
        :type positionalArgs: dict
        :param keywordArguments: keyword arguments of the function defined in self.func
        :type keywordArguments: dict

        :return: funcargs, pargs; These are keyword and positional arguments to self.func
        """
        pargs = []  # positional arguments
        funcargs = {}  # Function arguments

        # update the positional args
        positionalArgsMap = self._initialise_args(positionalArgs)
        keywordArgsMap = self._initialise_args(keywordArguments)
        positionalArgsMap, keywordArgsMap, input_outputs, vparg, vkarg = (
            self._populate_arguments_with_graph_data(positionalArgsMap, keywordArgsMap))

        # Determine if we can use the *args and **kwargs we identified above
        if self.varargs or self.populate_arguments_without_signature:
            logger.debug("Adding remaining *args to pargs %s", vparg)
            pargs.extend(vparg)
        if self.varkw:
            logger.debug("Adding remaining **kwargs to funcargs: %s", vkarg)
            funcargs.update(vkarg)

        # Update any InputOutput ports that might have path names
        for arg in input_outputs:
            keywordArgsMap, positionalArgsMap = self._update_filepaths(
                positionalArgsMap, keywordArgsMap, arg)

        # Extract arg and values from pargs; we no longer need the metadata
        logger.debug("Updating funcargs with values from pargsDict: %s",
                     positionalArgsMap)

        tmpPargs = {argstr: argument.value for argstr, argument in
                    positionalArgsMap.items()}

        funcargs.update(tmpPargs)
        logger.debug("positionalArgsMap: %s", positionalArgsMap)
        # Mixin the values from named ports
        portargs = self._ports2args(positionalArgsMap, keywordArgsMap)

        # Update any InputOutput ports that might have path names defined from input port
        for arg in input_outputs:
            keywordArgsMap, positionalArgsMap = self._update_filepaths(
                positionalArgsMap, keywordArgsMap, arg)

        logger.debug("Updating funcargs with values from named ports %s", portargs)
        tmpPortArgs = {port: arg.value for port, arg in portargs.items()}
        funcargs.update(tmpPortArgs)

        return funcargs, pargs

    def _update_filepaths(
            self,
            positionalArgsMap: dict[str, Argument],
            keywordArgsMap: dict[str, Argument],
            arg: str
    ):
        """
        Map any attribute that is an InputOutput

        This is used to allow PyFunc apps that require a filename as input to refer to
        the output data that is linked to the AppDrop. This resolves issues with us
        hidden having side effects of functions saving files without DALiuGE being
        aware of them.

        Parameters
        ----------
        positionalArgsMap: map of positional argument strs (from function signature) to the
        Argument objects (which store the current state of the values to be passed to
        the function).
        keywordArgsMap: map of keyword argument strs (from function signature) to the
        Argument objects (which store the current state of the values to be passed to
        the function).

        Returns
        -------
        modified keywordArgsMap and positionalArgsMap
        """

        # match to output in the event that it is None
        # Get the outputs accordingly

        attr_uid_map = {}
        output_port_count = {}
        for output in self.parameters["outputs"]:
            for key, value in output.items():
                attr_uid_map[value] = key
                if key in output_port_count:
                    output_port_count[key] += 1
                else:
                    output_port_count[key] = 1

        for arg_map in [keywordArgsMap, positionalArgsMap]:
            if arg in arg_map:
                output_uid = attr_uid_map[arg]
                output_drop = next(
                    (drop for drop in self.outputs if drop.uid == output_uid),
                    None
                )
                argument = arg_map[arg]
                parser = (DropParser(argument.encoding))
                if parser == DropParser.PATH:
                    try:
                        argument.value = filepath_from_string(
                            argument.value, path_type=output_drop.path_type,
                            dirname=output_drop.dirname,
                            uid=output_drop.uid,
                            humanKey=output_drop.humanKey
                        )
                    except RuntimeError as e:
                        raise InvalidPathException(
                            "Path contains unset environment variable", e
                        ) from e
                    self._output_filepaths[output_uid] = argument.value
                arg_map[arg] = argument
                self.parameters[arg] = arg_map[arg].value
        return keywordArgsMap, positionalArgsMap

    def _ports2args(self, pargsDict, keyargsDict) -> dict:
        """
        Replace arguments with values from ports.

        Returns:
        --------
        portargs dictionary
        """
        portargs = {}
        # 3. replace default argument values with named input ports
        logger.debug("Mapping from _inputs: %s", self._inputs)
        logger.debug("Parameters: %s", self.parameters)
        if "input_parser" in self.parameters:
            self.input_parser = self.parameters["input_parser"]
        if "output_parser" in self.parameters:
            self.output_parser = self.parameters["output_parser"]
        if "inputs" in self.parameters and check_ports_dict(self.parameters["inputs"]):
            logger.debug("Mapping ports to inputs...")
            if self.fn_nargs == 0:
                check_len = len(self._inputs)
            else:
                check_len = min(
                    len(self._inputs),
                    self.fn_nargs + self.fn_nkw,
                )
            inputs_dict = collections.OrderedDict()
            for inport in self.parameters["inputs"]:
                key = list(inport.keys())[0]
                inputs_dict[key] = {"name": inport[key], "path": None, "drop":
                    self._inputs[key]}
            parser = (
                get_port_reader_function(self.input_parser)
                if hasattr(self, "input_parser")
                else None
            )
            keyPortArgs, _ = identify_named_ports(
                inputs_dict,
                pargsDict,
                keyargsDict,
                check_len=check_len,
                mode="inputs",
                addPositionalToKeyword=True,
                parser=parser
            )
            portargs.update(keyPortArgs)
        else:
            for i, input_drop in enumerate(self._inputs.values()):
                parser = (
                    get_port_reader_function(self.input_parser)
                    if hasattr(self, "input_parser")
                    else get_port_reader_function("dill")
                )
                value = parser(input_drop)
                if self.argnames:
                    logger.debug("Port value pair: %s, %s",
                                 self.argnames[i],
                                 value)

        logger.debug(
            "Finally port mapping: %s, %s, %s", portargs, pargsDict, keyargsDict
        )
        return portargs

    def initialize_with_func_code(self):
        """
        This function takes over if code is passed in through an argument.
        """
        serialized = False
        logger.debug("Initializing with func_code of type %s",
                     type(self.func_code))
        if not isinstance(self.func_code, bytes):
            try:
                self.func = import_using_code(self.func_code, self.func_name,
                                              serialized=False)
            except (SyntaxError, NameError):
                serialized = True
        if isinstance(self.func_code, bytes) or serialized:
            if isinstance(self.func_code, str):
                self.func_code = base64.b64decode(self.func_code.encode("utf8"))
            self.func = import_using_code_ser(self.func_code, self.func_name)

        self._init_fn_defaults()
        # make sure defaults are dicts
        self._mixin_func_defaults()
        if isinstance(self.func_arg_mapping, str):
            self.func_arg_mapping = ast.literal_eval(self.func_arg_mapping)
        self.func_name = self.func.__qualname__
        return

    @track_current_drop
    def initialize(self, **kwargs):
        """
        The initialization of a function component is mainly dealing with mapping
        inputs and provided applicationArgs to the function arguments. All of this
        should be driven by matching names.
        """
        BarrierAppDROP.initialize(self, **kwargs)

        env = os.environ.copy()
        env.update({"DLG_UID": self._uid})
        if self._dlg_session_id:
            env.update({"DLG_SESSION_ID": self._dlg_session_id})
        elif "dlg_session_id" in kwargs:
            self._dlg_session_id = kwargs["dlg_session_id"]
            env.update({"DLG_SESSION_ID": self._dlg_session_id})

        self._applicationArgs = self._popArg(kwargs, "applicationArgs", {})

        self.func_code = self._popArg(kwargs, "func_code", None)
        self.func = self._popArg(kwargs, "func", None)
        self.arguments_defaults = []

        # backwards compatibility
        if "pickle" in self._applicationArgs:
            if self._applicationArgs["pickle"]["value"] == True:
                self.input_parser = DropParser.PICKLE
                self.output_parser = DropParser.PICKLE
            else:
                self.input_parser = DropParser.EVAL
                self.output_parser = DropParser.EVAL
            self._applicationArgs.pop("pickle")

        self._clean_applicationArgs()

        self.num_args = len(
            self._applicationArgs)  # number of additional arguments provided

        if not self.func_name and not self.func_code and not self.func:
            raise IncompleteDROPSpec(
                self, "No function specified (either via name, code or function object)"
            )
        if self.func:
            self.func_name = self.func.__name__
        self.name = f"{self.name}:{self.func_name}"  # PyFuncApp is parent
        logger.info("AppDROP name: %s", self.name)
        # Lookup function or import bytecode as a function
        if inspect.isfunction(self.func):
            self.argnames = list(inspect.signature(self.func).parameters)
            self._init_fn_defaults()
        elif not self.func and not self.func_code:
            self.func = import_using_name(self, self.func_name, curr_depth=0)
            self._init_fn_defaults()
        else:
            self.initialize_with_func_code()

        logger.debug("Args summary for %s", self.func_name)
        logger.debug("Args: %s", self.argnames)
        logger.debug("Args defaults:  %s", self.fn_defaults)
        logger.debug("Args pos/kw: %s", list(self.poskw.keys()))
        logger.debug("Args keyword only: %s", list(self.kwonly.keys()))
        logger.debug("VarArgs allowed:  %s", self.varargs)
        logger.debug("VarKwds allowed:  %s", self.varkw)

        # Mapping between argument name and input drop uids
        logger.debug("Input mapping provided: %s", self.func_arg_mapping)
        self._output_filepaths = {}
        self._recompute_data = {}

    @track_current_drop
    def run(self):
        """
        Function positional and keyword argument treatment:

        Function arguments can be provided in three different ways:
        1) Through an input port
        2) By specifying ApplicationArgs (one for each argument)
        3) Through defaults at the time of function definition

        The priority follows the list above with input ports overruling the others.
        Function arguments in Python can be passed as positional, kw-value, positional
        only, kw-value only, and catch-all args and kwargs, which don't provide any
        hint about the names of accepted parameters. All of them are now supported. If
        positional arguments or kw-value arguments are provided by the user, but are
        not explicitely defined in the function signiture AND args and/or kwargs are
        allowed then these arguments are passed to the function. For args this is
        somewhat risky, since the order is relevant and in this code derived from the
        order defined in the graph (same order as defined in the component description).

        Input ports will NOT be used by index (anymore), but by using the name of the port.
        Since each input port requires an associated data drop, this provides a unique
        mapping. This also allows to pass values to any function argument through a port.

        """
        self._run()
        logger.user("Confirm the operator exception works.")
        logger.debug("This object: %s, %s", self, self._humanKey)
        funcargs = {}
        # Keyword arguments are made up of the default values plus the inputs
        # that match one of the keyword argument names
        # if defaults dict has not been specified at all we'll go ahead anyway

        # 1. Fill arguments with rest of inputs
        # fill the pargsDict with positional and poskw arguments and defaults
        pargsDict = {k: v.default for k, v in self.posonly.items()}
        pargsDict.update({k: v.default for k, v in self.poskw.items()})
        logger.debug("Initial pos_kwargs dictionary: %s", pargsDict)
        # fill the keyargsDict with kwonly arguments and defaults
        keyargsDict = {k: v.default for k, v in self.kwonly.items()}
        logger.debug("Initial kwonly dictionary: %s", self.kwonly)

        for arg in pargsDict:
            pargsDict[arg] = self.fn_defaults[arg]
        for arg in keyargsDict:
            keyargsDict[arg] = self.fn_defaults[arg]

        # 2. Map arguments from user-defined data, or from InputPorts
        funcargs, pargs = self._map_parameters_to_func_args(pargsDict, keyargsDict)

        self._recompute_data["args"] = funcargs.copy()

        # 3. remove self argument if this is the initializer.
        if (self.func_name is not None
                and self.func_name.split(".")[-1] in ["__init__", "__class__"]
                and "self" in funcargs):
            funcargs.pop("self")

        logger.user("Running %s", self.func_name)
        logger.user("Arguments: *%s **%s", pargs, funcargs)

        # 4. prepare for execution
        # we capture and log whatever is produced on STDOUT
        capture = StringIO()
        try:
            bind = self.argsig.bind(*pargs, **funcargs)
            logger.debug("Bound arguments: %s", bind)
        except TypeError as e:
            logger.error("Binding of arguments failed: %s", e)
        except AttributeError:
            logger.debug("Binding failed due to signature not being callable")

        # 5. Here is where the function is actually executed
        with redirect_stdout(capture):
            if isinstance(self.argsig, inspect.Signature):
                self.result = self.func(*bind.args, **bind.kwargs)
            else:
                self.result = self.func(*pargs, **funcargs)

        logger.user("Returned result from %s: %s", self.func_name, self.result)
        if capture.getvalue():
            msg = f"STDOUT/STDERR output from function: '{self.func_name}': {capture.getvalue()}"
        else:
            msg = f"No STDOUT/STDERR output from function: '{self.func_name}'"
        logger.user(msg)
        logger.debug("Finished execution of %s", self.func_name)

        # 6. Process results
        # Depending on how many outputs we have we treat our result
        # as an iterable or as a single object. Each result is pickled
        # and written to its corresponding output
        self.write_results()

    def _match_parser(self, output_drop):
        """
        Match the output parser to the appropriate drop
        """

        encoding = "dill"
        # if not self._applicationArgs:
        #     return getattr(self, "output_parser", "dill")
        component_params = self.parameters.get("componentParams", {})
        applicationArgs = self.parameters.get("applicationArgs", {})
        if "outputs" in self.parameters and check_ports_dict(
                self.parameters["outputs"]
        ):
            for outport in self.parameters["outputs"]:
                drop_uid, drop_port = list(outport.items())[0]
                if drop_uid == output_drop.uid and drop_port in applicationArgs:
                    param_enc = applicationArgs[drop_port]["encoding"]
                    encoding = param_enc or encoding
                elif drop_uid == output_drop.uid and drop_port in component_params:
                    param_enc = component_params[drop_port]["encoding"]
                    encoding = param_enc or encoding
        return DropParser(encoding) if encoding else self.output_parser

    def write_results(self):
        from dlg.droputils import listify

        result_iter = listify(self.result)
        if not self.outputs and result_iter:
            return

        logger.debug(
            "Writing following result to %d outputs: %s", len(self.outputs), result_iter
        )
        num_outputs = len(self.outputs)
        for i, o in enumerate(self.outputs):
            # Ensure that we don't produce two files for the same output DROP
            if o.uid in self._output_filepaths:
                # Trigger FileDROP filename update, but don't write to the drop because
                # it has already been written to.
                _ = o.getIO()
                # 'Discount the Filepath output from outputs we write to'
                num_outputs -= 1
                continue
            if not result_iter and self.outputs:
                result = result_iter
            elif len(result_iter) == 1:
                # We only have one element, no need to save as a list
                result = result_iter[0]
            elif len(result_iter) > 1 and num_outputs == 1:
                # We want all elements in the list to go to the output
                result = self.result
            else:
                # Iterate over each element of the list for each output
                # Wrap around for len(result_iter) < len(self.outputs)
                i = i % len(self.outputs)
                result = result_iter[i]

            parser = self._match_parser(o)
            parser = resolve_drop_parser(parser)
            if parser is DropParser.PICKLE:
                o.write(pickle.dumps(result))
            elif parser is DropParser.DILL:
                logger.debug("Writing dilled result %s to %s", type(result), o)
                o.write(dill.dumps(result))
            elif parser is DropParser.EVAL or parser is DropParser.UTF8:
                if isinstance(result, str):
                    o.write(result)
                else:
                    encoded_result = repr(result).encode("utf-8")
                    o.write(encoded_result)
            elif parser is DropParser.NPY:
                import numpy as np

                if not isinstance(result, np.ndarray):
                    try:
                        result = np.array(result)
                    except Exception as e:
                        raise (e)
                drop_loaders.save_npy(o, result)
            elif parser is DropParser.RAW:
                o.write(result)
            elif parser is DropParser.BINARY:
                drop_loaders.save_binary(o, result)
            else:
                raise ValueError(parser.__repr__())

    def generate_recompute_data(self):
        for name, val in self._recompute_data.items():
            try:
                json.dumps(val)
            except TypeError as e:
                logger.debug(e)
                self._recompute_data[name] = repr(val)
        return self._recompute_data



# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '6.0.0'
git_version = 'd266e83f3d79f1413c264ae387c41e81097edbc4'
full_version = '6.0.0'
is_release = True

if not is_release:
    version = full_version


#
#    ICRAR - International Centre for Radio Astronomy Research 2024
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA

import threading
import time
import datetime

from dlg.drop import track_current_drop
from dlg.apps.app_base import BarrierAppDROP
from dlg.common import get_roots
from dlg.data.drops.data_base import logger
from dlg.ddap_protocol import AppDROPStates, DROPStates
from dlg.dropmake.pg_generator import unroll, partition
from dlg.manager.composite_manager import DataIslandManager
from dlg.manager.node_manager import NodeManager
from dlg.manager.drop_manager import DROPManager
from dlg.manager.rest import NMRestServer, CompositeManagerRestServer


def shutdownManager(managers: dict) -> None:
    """
    Helper function that iterates through a dictionary of managers and stops them and
    their respective servers.
    :param managers:dict of dicts, which contain DROPManager objects and information.
    :return:
    """
    for manager in managers.values():
        manager["manager"].shutdown()
        manager["server"].stop()


def startupManagersInThread(managers: dict) -> None:
    """
    Helper function that iterates through a dictionary of managers and starts
    them in separate threads.
    :param managers: dict of dicts, which contain DROPManager objects and information.
    :return: None
    """
    for manager in managers.values():
        thread = threading.Thread(
            target=manager["server"].start, args=(manager["host"], manager["port"])
        )
        thread.start()


##
# @brief SubGraphLocal
# @par EAGLE_START
# @param category DALiuGEApp
# @param tag daliuge
# @param log_level "NOTSET"/Select/ComponentParameter/NoPort/ReadWrite/NOTSET,DEBUG,INFO,WARNING,ERROR,CRITICAL/False/False/Set the log level for this drop
# @param dropclass dlg.apps.subgraph.SubGraphLocal/String/ComponentParameter/NoPort/ReadOnly//False/False/
# @param base_name subgraph/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param execution_time 5/Float/ConstraintParameter/NoPort/ReadOnly//False/False/Estimated execution time
# @param num_cpus 1/Integer/ConstraintParameter/NoPort/ReadOnly//False/False/Number of cores used
# @param group_start True/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the start of a group?
# @param node_manager_port 8080/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/The port number of the SubGraph Node Manager
# @param island_manager_port 8081/Integer/ApplicationArgument/NoPort/ReadWrite//False/False/The port number of the SubGraph Island Manager
# @param node_manager_host localhost/string/ApplicationArgument/NoPort/ReadWrite//False/False/The node manager host address
# @param island_manager_host localhost/string/ApplicationArgument/NoPort/ReadWrite//False/False/The island manager host address
# @param partition_algorithm metis/string/ApplicationArgument/NoPort/ReadWrite//False/False/The island manager host address
# @par EAGLE_END
class SubGraphLocal(BarrierAppDROP):
    """
    InputApp for the SubGraph construct, designed to be used on a system with DALiuGE
    installed locally.
    """

    DEFAULT_SUBGRAPH_NODE_PORT = 8080
    DEFAULT_SUBGRAPH_ISLAND_PORT = 8081

    def initialize(self, **kwargs):
        super(SubGraphLocal, self).initialize(**kwargs)
        self._nm_port = self._popArg(
            kwargs, "node_manager_port", self.DEFAULT_SUBGRAPH_NODE_PORT
        )
        self._dim_port = self._popArg(
            kwargs, "island_manager_port", self.DEFAULT_SUBGRAPH_ISLAND_PORT
        )
        self._subgraph = self._popArg(kwargs, "subgraph", {})

        # List of nodes (IP's, str) that are used in
        self._nodeManagerHost = self._popArg(
            kwargs, "node_manager", "localhost"
        )  # :8080"])
        self._islandManagerHost = self._popArg(kwargs, "island_manager", "localhost")
        self._partition_algorithm = self._popArg(kwargs, "partition_algorithm", "metis")
        self._session_prefix = self._popArg(kwargs, "session_prefix", "subgraph")

    def _pollRESTInterface(
        self, manager: DROPManager, session: str, pollFrequencyInSeconds: int = 30
    ):
        """
        Check the completion status of each drop in the SubGraph.

        If not all have completed, sleep before returning the status

        :param manager: DropManager, the manager that is deploying the subgraph
        :param pollFrequencyInSeconds: number of times we poll the manager for
        :return:
        """
        finished = all(
            status["status"] in [DROPStates.ERROR, DROPStates.COMPLETED]
            for host, status in manager.getGraphStatus(session).items()
        )
        if not finished:
            time.sleep(pollFrequencyInSeconds)
        return finished

    def _translateSubGraph(self, node_list: list = None):
        """
        :param node_list: list of both data island and node managers
        ------

        Use a local install of the daliuge-translator to convert the LGT into a
        partitioned pg_spec.
        """
        if not list:
            return {}
        pgt = unroll(self._subgraph)
        pg = partition(
            pgt,
            algo="metis",
            num_partitions=1,
            num_islands=1,
            partition_label="Partition",
            show_gojs=True,
            max_cpu=8,
            max_load_imb=100,
        )
        return pg.to_pg_spec(node_list, ret_str=False)

    @track_current_drop
    def run(self):
        """
        Start the required DROPManagers and deploy the subgraph

        This method will wait until confirmation from the DROPManager that the DROPs have
        either successfully completed or failed.
        """
        managers = {
            "node": {
                "manager": NodeManager(
                    self._nodeManagerHost, rpc_port=6667, events_port=5556
                ),
                "port": self._nm_port,
                "host": self._nodeManagerHost,
            },
            "dim": {
                "manager": DataIslandManager([self._nodeManagerHost]),
                "port": self._dim_port,
                "host": self._islandManagerHost,
            },
        }
        managers["node"]["server"] = NMRestServer(managers["node"])
        managers["dim"]["server"] = CompositeManagerRestServer(managers["dim"])
        startupManagersInThread(managers)

        nodes = [self._islandManagerHost, self._nodeManagerHost]
        try:
            session = self._translateAndDeploySubGraph(
                managers["dim"]["manager"], nodes
            )
            while not self._pollRESTInterface(
                manager=managers["dim"]["manager"],
                session=session,
                pollFrequencyInSeconds=10,
            ):
                logger.info("Running SubGraph externally")

        except RuntimeError as e:
            logger.debug("Exception when deploying subgraph: %s", e)
            shutdownManager(managers)
            self.execStatus = AppDROPStates.CANCELLED
            self.status = DROPStates.CANCELLED
            return self.status

        logger.info("Subgraph finished executing")
        shutdownManager(managers)
        return self.status

    def _translateAndDeploySubGraph(self, dim: DROPManager, nodes: list) -> str:
        """
        Translate the subgraph and prepare it for deployment.

        :param dim: Data Island Manager
        :param nodes: List of DropManagers used (DIM + NM)
        :return: None
        """
        session_id = self._generate_session_id()
        graphspec = self._translateSubGraph(nodes)
        dim.createSession(session_id)
        dim.addGraphSpec(session_id, graphspec)
        # Roots are required to 'kick off' the drop execution
        roots = get_roots(graphspec)
        dim.deploySession(session_id, completedDrops=roots)
        return session_id

    def _generate_session_id(self) -> str:
        """
        Create a session_id using a prefix + timestamp format

        :param prefix: str, an optionally-configurat session ID prefix
        :return: A string prefix
        """
        ts = time.time()
        return f"{self._session_prefix}_" + datetime.datetime.fromtimestamp(
            ts
        ).strftime("%Y-%m-%d_%H-%M-%S")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
from unittest.case import skipIf

# pylint: disable=possibly-used-before-assignment
"""
Test the S3 Drop
"""
# If the profile is not present in the current user's account
# we simply skip the test

import unittest

run_tests = True
try:
    import boto3
    from dlg.data.drops.s3_drop import S3DROP
except ImportError:
    run_tests = False

if run_tests:
    PROFILE = "acacia-awicenec"
    try:
        boto3.session.Session(profile_name=PROFILE)
        bucket = "rascil"
        key = "2022-09-27T17:30:02_-13_0/1"
        endpoint_url = "https://projects.pawsey.org.au"
    except:
        run_tests = False


class TestS3Drop(unittest.TestCase):
    @skipIf(not run_tests, "No profile found to run this test")
    def test_bucket_exists(self):
        drop = S3DROP(
            "oid:A", "uid:A", 
            profile_name=PROFILE, 
            Bucket="ska-low-sim", 
            Key="Nonsense",
            endpoint_url=endpoint_url,
        )
        self.assertEqual(drop.exists(), False)
        del(drop)

        drop = S3DROP(
            "oid:A",
            "uid:A",
            profile_name=PROFILE,
            Bucket=bucket,
            Key=key,
            endpoint_url=endpoint_url,
        )
        self.assertEqual(drop.exists(), True)

    @skipIf(not run_tests, "No profile found to run this test")
    def test_size(self):
        drop = S3DROP(
            "oid:A", "uid:A", 
            profile_name=PROFILE, 
            Bucket="ska-low-sim", 
            Key="Nonsense",
            endpoint_url=endpoint_url,
        )
        self.assertNotEqual(drop.size, 2785011)
        del(drop)

        drop = S3DROP('oid:A', 'uid:A', 
            profile_name=PROFILE, 
            Bucket=bucket, 
            Key=key,
            endpoint_url=endpoint_url,
            )
        self.assertEqual(drop.size, 2785011)
    
    @skipIf(not run_tests, "No profile found to run this test")
    def test_read(self):
        
        drop = S3DROP('oid:A', 'uid:A', 
            profile_name=PROFILE, 
            Bucket=bucket, 
            Key=key,
            endpoint_url=endpoint_url,
            )
        drop.status = 2
        desc = drop.open()
        self.assertEqual(len(drop.read(desc, count=-1)), 2785011)

    @skipIf(not run_tests, "No profile found to run this test")
    def test_write(self):
        """
        This is also testing delete.
        """
        drop = S3DROP('oid:A', 'uid:A', 
            profile_name=PROFILE, 
            Bucket=bucket, 
            Key="testData",
            endpoint_url=endpoint_url,
            )
        if drop.exists():
            drop.delete()
        testdata = 10000*b"a"
        written = drop.write(testdata)
        self.assertEqual(written, len(testdata))
        drop.delete()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Generate the SLURM script, and submit it to the queue based on various paramters

parse the log result, and produce the plot

"""

import json
import optparse # pylint: disable=deprecated-module
import sys
import os

from pathlib import Path

from dlg.deploy.configs import (
    ConfigFactory,
)  # get all available configurations

from dlg.deploy.configs import (
    DEFAULT_MON_PORT,
    DEFAULT_MON_HOST,
)
from dlg.deploy.logparser import LogParser
from dlg.deploy.slurm_client import SlurmClient
from dlg.common.reproducibility.reproducibility import (
    init_pgt_unroll_repro_data,
    init_pgt_partition_repro_data,
)

from dlg.deploy.configs.config_manager import ConfigManager, ConfigType

from dlg.dropmake import pg_generator

FACILITIES = ConfigFactory.available()

def process_bool(param: str):
    """
    Evaluate the boolean value from a string, given bool('False') is True.

    :param param: The parameter string we want to verify
    :return:
    """
    if isinstance(param, bool):
        return param

    if param.lower() == 'false':
        return False
    elif param.lower() == 'true':
        return True
    else:
        raise ValueError("You have likely misspelled True/False in your .ini config")

def process_config(config_file: str):
    """
    Use configparser to process INI file

    Current functionality: 
        - Returns remote environment config (e.g. DLG_ROOT, HOME etc.)

    Future Functionality: 
        - Graph translation parameters
        - Engine parameters

    :returns: dict, config information
    """
    from configparser import ConfigParser, ExtendedInterpolation
    parser = ConfigParser(interpolation=ExtendedInterpolation())
    all_opts = {}
    parser.read(config_file)
    for s in parser.sections():
        all_opts |= (dict(parser[s]))
    return {key: value for key, value in all_opts.items() if value}

def process_slurm_template(template_file: str):
    template = Path(template_file)
    with template.open('r') as fp:
        return fp.read()

def create_engine_group(parser: optparse.OptionParser):
    """
    Setup optparse group for all DALiuGE Engine options.

    This includes number of Node and Data-Island managers , logging levels, and
    verbosity settings.

    :param parser:
    :return: parser
    """
    group = optparse.OptionGroup(parser, "Engine options",
                                 "DALiuGE engine configuration and runtime options")


    group.add_option(
        "-l",
        "--log-root",
        action="store",
        dest="log_root",
        help="The root directory of the log file",
    )

    group.add_option(
        "-d",
        "--log-dir",
        action="store",
        dest="log_dir",
        help="The directory of the log file for parsing",
    )
    group.add_option(
        # TODO: num_nodes needs to be >= #partitions in PGT, if provided
        "-n",
        "--num_nodes",
        action="store",
        type="int",
        dest="num_nodes",
        help="Number of compute nodes requested",
    )

    group.add_option(
        "-s",
        "--num_islands",
        action="store",
        type="int",
        dest="num_islands",
        help="The number of Data Islands",
    )

    group.add_option(
        "-u",
        "--all_nics",
        action="store_true",
        dest="all_nics",
        help="Listen on all NICs for a node manager",
        default=False,
    )

    group.add_option(
        "-S",
        "--check_with_session",
        action="store_true",
        dest="check_with_session",
        help="Check for node managers' availability by creating/destroy a session",
        default=False,
    )

    group.add_option(
        "-D",
        "--dlg_root",
        dest="dlg_root",
        action="store",
        type="string",
        help="Overwrite the DLG_ROOT directory provided by the config",
    )

    group.add_option(
        "-v",
        "--verbose-level",
        action="store",
        type="int",
        dest="verbose_level",
        help="Verbosity level (1-3) of the DIM/NM logging",
        default=1,
    )

    parser.add_option(
        "-c",
        "--csvoutput",
        action="store",
        dest="csv_output",
        help="CSV output file to keep the log analysis result",
    )

    return group

def create_slurm_group(parser: optparse.OptionParser):
    group = optparse.OptionGroup(parser, "Slurm options",
                                 "Slurm job script options. "
                                 "Note: These will be overwritten if using "
                                 "--slurm_template")
    group.add_option(
        "-T",
    "--max-threads",
    action="store",
    type="int",
    dest="max_threads",
    help="Max thread pool size used for executing drops. 0 (default) means no pool.",
    default=0,
    )

    group.add_option(
    "-f",
    "--facility",
    dest="facility",
    choices=FACILITIES,
    action="store",
    help=f"The facility for which to create a submission job\nValid options: {FACILITIES}",
    default=None,
    )

    group.add_option(
        "-t",
        "--job-dur",
        action="store",
        type="int",
        dest="job_dur",
        help="job duration in minutes",
        default=30,
    )

    return group

def create_config_group(parser: optparse.OptionParser):
    """
    Establish configuration group for config-file and template based options.

    :param parser: parser that we are updating
    :return: the group for experiments
    """
    group=optparse.OptionGroup(parser, "Remote configuration Options",
                      "Remote deployment configuration options based on configuration "
                      "and template files.")

    group.add_option(
        "--config_file",
        dest="config_file",
        type="string",
        action="store",
        help="Use INI configuration file.",
        default=None
    )
    group.add_option(
        "--slurm_template",
        dest="slurm_template",
        type="string",
        action="store",
        help="Use SLURM template file for job submission. WARNING: Using this command will over-write other job-parameters passed here.",
        default=None
    )

    return group

def create_local_graph_group(parser):
    """
    TODO: LIU-424
    """
    group = optparse.OptionGroup(parser, "Local graph options",
                      "Options for locally stored graphs")
    group.add_option(
        "-L",
        "--logical-graph",
        action="store",
        type="string",
        dest="logical_graph",
        help="The filename of the logical graph to deploy",
        default=None,
    )

    group.add_option(
        "-P",
        "--physical-graph",
        action="store",
        type="string",
        dest="physical_graph",
        help="The filename of the physical graph (template) to deploy",
        default=None,
    )

    return group


def create_remote_graph_group(parser):
    """
    TODO: LIU-424
    """
    group = optparse.OptionGroup(parser, "Remote graph options",
                      "Options for graphs stored in remote repositories. "
                               "Currently supported: GitHub, GitLab")

    group.add_option(
        "--github",
        action="store_true",
        dest="github",
        help="Access graph from remote repository",
        default=False,)

    group.add_option("--gitlab",
        action="store_true",
        dest="gitlab",
        help="Access graph from remote repository",
        default=False,)

    group.add_option("--user_org")
    group.add_option("--repo")
    group.add_option("--branch")
    group.add_option("--path")

    return group


def evaluate_graph_options(opts, parser):
    """
    Perform checks on the graph options to ensure mutually exclusive information is not
    stored.
    :param opts:
    :return:
    """

    from dlg.deploy.remote_graph import github_request, gitlab_request

    use_remote_graph = opts.github or opts.gitlab
    use_local_graph = opts.logical_graph or opts.physical_graph

    graph = None
    if use_local_graph:
        if use_remote_graph:
            parser.error("Cannot specify both local and remote graph")
        else:
            if opts.logical_graph:
                graph =  _translate_graph(parser, opts.logical_graph)
            if opts.physical_graph:
                graph =  opts.physical_graph

    if opts.github:
        content = github_request(opts.user_org, opts.repo, opts.branch, opts.path)
        graph = _translate_graph(parser, opts, content)
    if opts.gitlab:
        content = gitlab_request(opts.user_org, opts.repo, opts.branch, opts.path)
        graph =  _translate_graph(parser, opts, content)

    graph = 'mytest.graph'
    return graph


def create_monitor_options(parser):
    group = optparse.OptionGroup(parser, "Monitor proxy options",
                                 "Start and configure the monitoring proxy.")


    group.add_option(
        "-p",
        "--run_proxy",
        action="store_true",
        dest="run_proxy",
        help="Whether to attach proxy server for real-time monitoring",
        default=False,
    )
    group.add_option(
        "-m",
        "--monitor_host",
        action="store",
        type="string",
        dest="mon_host",
        help="Monitor host IP (optional)",
        default=DEFAULT_MON_HOST,
    )
    group.add_option(
        "-o",
        "--monitor_port",
        action="store",
        type="int",
        dest="mon_port",
        help="The port to bind DALiuGE monitor",
        default=DEFAULT_MON_PORT,
    )

    group.add_option(
        "-i",
        "--visualise_graph",
        action="store_true",
        dest="visualise_graph",
        help="Whether to visualise graph (poll status)",
        default=False,
    )
    return group


def create_component_options(parser):
    group = optparse.OptionGroup(parser, "Graph Component options",
                                 "Update component DROPs for testing.")

    group.add_option(
        "-z",
        "--zerorun",
        action="store_true",
        dest="zerorun",
        help="Generate a physical graph that takes no time to run",
        default=False,
    )
    group.add_option(
        "-y",
        "--sleepncopy",
        action="store_true",
        dest="sleepncopy",
        help="Whether include COPY in the default Component drop",
        default=False,
    )
    return group

def create_algorithm_options(parser):
    group = optparse.OptionGroup(parser, "Algorithm options",
                                 "Algorithm choices for translation and "
                                 "partitioning.")

    # Algorithm options
    group.add_option(
        "-A",
        "--algorithm",
        action="store",
        type="string",
        dest="algorithm",
        help="The algorithm to be used for the translation",
        default="metis",
    )

    group.add_option(
        "-O",
        "--algorithm-parameters",
        action="store",
        type="string",
        dest="algorithm_params",
        help="Parameters for the translation algorithm",
    )

    return group

def _translate_graph(parser, opts, content=""):
    if opts.logical_graph:
        graph_file = os.path.basename(opts.logical_graph)
    else:
        graph_file = opts.path.split("/")[-1]
    pre, ext = os.path.splitext(graph_file)
    if os.path.splitext(pre)[-1] != ".pre":
        pgt_file = ".".join([pre, "pgt", ext[1:]])
    else:
        pgt_file = graph_file

    lg_graph = {}
    if opts.logical_graph:
        with open(opts.logical_graph) as f:
            # logical graph provided, translate first
            lg_graph = json.loads(f.read())
    elif content:
        lg_graph = json.loads(content)

    else:
        parser.error("Incorrect configuration, no graph available to translate")
        sys.exit(1)
    pgt = pg_generator.unroll(lg_graph, zerorun=opts.zerorun)
    pgt = init_pgt_unroll_repro_data(pgt)
    reprodata = pgt.pop()
    pgt = pg_generator.partition(
        pgt=pgt,
        algo=opts.algorithm,
        algo_params=opts.algorithm_params,
        num_islands=int(opts.num_islands),
        num_partitions=int(opts.num_nodes),
    )
    pgt.append(reprodata)
    pgt = init_pgt_partition_repro_data(pgt)
    pgt_name = pgt_file
    pgt_path = Path(f"/tmp/{pgt_file}")
    with pgt_path.open("w") as o:
        json.dump((pgt_name, pgt), o, indent=2)
        return str(pgt_path)


def analyse(opts, parser):
    """
    Run the analysis procedure

    :param opts: the optparse Values
    :param parser: the optparse Parser that we use for error reporting
    """

    if opts.log_dir is None:
        # you can specify:
        # either a single directory
        if opts.log_root is None:
            config = ConfigFactory.create_config(
                facility=opts.facility, user=opts.username
            )
            log_root = config.getpar("log_root")
        else:
            log_root = opts.log_root
        if log_root is None or (not os.path.exists(log_root)):
            parser.error(
                "Missing or invalid log directory/facility for log analysis"
            )
        # or a root log directory
        else:
            for log_dir in os.listdir(log_root):
                log_dir = os.path.join(log_root, log_dir)
                if os.path.isdir(log_dir):
                    try:
                        log_parser = LogParser(log_dir)
                        log_parser.parse(out_csv=opts.csv_output)
                    except Exception as exp:  # pylint: disable=broad-exception-caught
                        parser.error(
                            "Fail to parse {0}: {1}".format(log_dir, exp))
    else:
        log_parser = LogParser(opts.log_dir)
        log_parser.parse(out_csv=opts.csv_output)


def submit(opts, parser):
    """
    Run the submit procedure to a remote facility

    :param opts: the optparse Values
    :param parser: the optparse Parser that we use for error reporting
    """

    cfg_manager = ConfigManager(FACILITIES)

    if opts.config_file:
        config_path = cfg_manager.load_user_config(ConfigType.ENV, opts.config_file)
        if not config_path:
            parser.error("Provided --config_file option that does not exist!")
            sys.exit(1)
        config = process_config(str(config_path)) if config_path else {}
        for attr, val in config.items():
            setattr(opts, attr, val)
    else:
        config = None

    if opts.github and opts.gitlab:
        parser.error("Must specify only one of --github or gitlab")

    pgt_file = evaluate_graph_options(opts, parser)
    if not pgt_file or not Path(pgt_file).exists():
        parser.error("There was an issue translating the physical graph and no graph "
                     "file could be found!")
        sys.exit(1)
    if opts.slurm_template:
        template_path = cfg_manager.load_user_config(ConfigType.SLURM, opts.slurm_template)
        if not template_path:
            parser.error("Provided --slurm_template option that does not exist!")
            sys.exit(1)
        template = process_slurm_template(template_path)  if template_path else None
    else:
        template = None

    client = SlurmClient(
        dlg_root=opts.dlg_root,
        facility=opts.facility,
        job_dur=int(opts.job_dur),
        num_nodes=int(opts.num_nodes),
        logv=int(opts.verbose_level),
        zerorun=opts.zerorun,
        max_threads=int(opts.max_threads),
        run_proxy=opts.run_proxy,
        mon_host=opts.mon_host,
        mon_port=opts.mon_port,
        num_islands=int(opts.num_islands),
        all_nics=process_bool(opts.all_nics),
        check_with_session=opts.check_with_session,
        physical_graph_template_file=pgt_file,
        submit=process_bool(opts.submit),
        remote=process_bool(opts.remote),
        username=opts.username,
        ssh_key=opts.ssh_key,
        config=config,
        slurm_template=template
    )

    client.visualise_graph = opts.visualise_graph
    client.submit_job()


def run(_, args):
    parser = optparse.OptionParser(
        usage="\n%prog --action [submit|analyse] -f <facility> [options]\n\n%prog -h for further help"
    )

    parser.add_option(
        "-a",
        "--action",
        action="store",
        type="choice",
        choices=["submit", "analyse"],
        dest="action",
        help="**submit** job or **analyse** log",
    )
    parser.add_option(
        "--submit",
        dest="submit",
        action="store_true",
        help="If set to False, the job is not submitted, but the script is generated",
        default=False,
    )

    parser.add_option(
        "--remote",
        dest="remote",
        action="store_true",
        help="If set to True, the job is submitted/created for a remote submission",
        default=False,
    )

    parser.add_option_group(create_algorithm_options(parser))
    parser.add_option_group(create_engine_group(parser))
    parser.add_option_group(create_slurm_group(parser))
    parser.add_option_group(create_monitor_options(parser))
    parser.add_option_group(create_component_options(parser))
    parser.add_option_group(create_remote_graph_group(parser))
    parser.add_option_group(create_local_graph_group(parser))
    parser.add_option_group(create_config_group(parser))

    # SSH options
    parser.add_option(
        "-U",
        "--username",
        dest="username",
        type="string",
        action="store",
        help="Remote username, if different from local",
        default=None,
    )

    parser.add_option(
        "--ssh_key",
        action="store",
        help="Path to ssh private key",
        default=None
    )

    (opts, _) = parser.parse_args(sys.argv)
    if len(sys.argv) <=1:
        parser.print_help()
        sys.exit(0)

    facility_necessary = False if opts.config_file else True

    if not opts.action and (facility_necessary and not opts.facility):
        parser.error("Missing required parameters!")
        parser.print_help()
        sys.exit(1)
    if facility_necessary and opts.facility not in FACILITIES:
        parser.error(f"Unknown facility provided. Please choose from {FACILITIES}")
        sys.exit(1)

    if opts.action == "analyse":
        analyse(opts, parser)
    elif opts.action == "submit":
        submit(opts, parser)
    else:
        parser.print_help()
        parser.error(f"Invalid input from args: {args}!")

if __name__ == "__main__":
    run(None, sys.argv[1:])



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import logging
import os
import shutil

from dlg.data import path_builder
from dlg.data.drops.data_base import PathBasedDrop, DataDROP
from dlg.exceptions import InvalidDropException
from dlg.meta import dlg_bool_param
from dlg.data.io import DirectoryIO

logger = logging.getLogger(f"dlg.{__name__}")


##
# @brief Directory
# @details A DataDROP that represents a filesystem directory.
# @par EAGLE_START
# @param category Directory
# @param tag daliuge
# @param dropclass dlg.data.drops.directory.DirectoryDROP/String/ComponentParameter/NoPort/ReadWrite//False/False/Drop class
# @param base_name directory/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param check_exists False/Boolean/ApplicationArgument/NoPort/ReadWrite//False/False/Perform a check to make sure the file path exists before proceeding with the application
# @param dirname /String/ApplicationArgument/NoPort/ReadWrite//False/False/"Directory name/path"
# @param create_if_missing False/Boolean/ApplicationArgument/NoPort/ReadWrite//False/False/"Create directory if it does not exist"
# @param overwrite_existing False/Boolean/ApplicationArgument/NoPort/ReadWrite//False/False/"Overwrite existing directory if exists"
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param io /Object/ApplicationArgument/OutputPort/ReadWrite//False/False/Input Output port
# @par EAGLE_END
class DirectoryDROP(PathBasedDrop, DataDROP):
    """
    A DataDROP that represents a filesystem directory.

    This is used as a proxy for directories on the system, and does not automatically
    append an arbitrary filename to the directory if it does not exist.
    """

    check_exists = dlg_bool_param("check_exists", False)
    create_if_missing = dlg_bool_param("create_if_missing", False)

    def initialize(self, **kwargs):
        DataDROP.initialize(self, **kwargs)

        if "dirname" not in kwargs:
            raise InvalidDropException(
                self, 'DirectoryContainer needs a "dirname" parameter'
            )
        self.dirpath = os.path.expandvars(kwargs["dirname"])
        self.dirname = self.dirpath
        self._path_type = path_builder.PathType.Directory
        self._setupDirectoryPath()


    def getIO(self):
        """
        Return DirectoryIO object
        """
        if not self.parameters.get("dirname", None):
            self._map_input_ports_to_params()
            self._setupDirectoryPath()
        return DirectoryIO(self._path)

    def _setupDirectoryPath(self):
        """
        Do the same as file._setupFilePaths()
        :return:
        """

        logger.debug("Checking existence of %s %s", self.dirpath, self.check_exists)
        # if "check_exists" in kwargs and kwargs["check_exists"] is True:
        if self.check_exists:
            if not os.path.isdir(self.dirpath):
                raise InvalidDropException(self, f"{self.dirpath} is not a directory")
        if not self._path:
            if not self.dirpath: # If nothing is provided
                self.dirpath = path_builder.base_uid_pathname(self.uid, self._humanKey)
            self._path = self.get_dir(self.dirpath, self.create_if_missing)

        self.dirname = self._path


    def delete(self):
        shutil.rmtree(self._path)

    def exists(self):
        return os.path.isdir(self._path)

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]  # Dervied from FileDROP
        return "file://" + hostname + self._path



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import base64
import binascii
import builtins
from json import JSONDecodeError
from pickle import PickleError

import dill
import io
import json
import os
import random
import string
import sys
from typing import Union

from dlg.utils import serialize_data
from dlg.common.reproducibility.reproducibility import common_hash
from dlg.data.drops.data_base import DataDROP, logger
from dlg.data.io import SharedMemoryIO, MemoryIO
from dlg.drop import track_current_drop


def get_builtins()-> dict:
    """
    Get a tuple of buitlin types to compare pydata with.
    """
    builtin_types = tuple(getattr(builtins, t) for t in dir(builtins) if isinstance(getattr(builtins, t), type))
    builtin_types = builtin_types[builtin_types.index(bool):]
    builtin_names = [b.__name__ for b in builtin_types]
    return dict(zip(builtin_names, builtin_types))

def guess_type(value):
    """Guess the type from value

    Args:
        value (Any): _description_
    """
    try:
        vtype = type(value).__name__
        return "raw" if vtype == "bytes" else vtype
    except NameError:
        return "raw"

def parse_pydata(pd: Union[bytes, dict]) -> bytes:
    """
    Parse and evaluate the pydata argument to populate memory during initialization

    :param pd: either the pydata dictionary from the graph node or the value directly

    This function attempts to identify the type of data so it can appropriately be stored
    in the right IO buffer (currently supporting BytesIO and StringIO).

    If there is nothing in pydata (e.g. the empty string, ''), then technically the
    type of the input data is String. However, if this is because the memory drop is
    expecting data from an InputPort, then this is not a good assumption.

    :returns a byte encoded value
    """
    stypes = ["float", "int", "str", "list", "dict"]
    if not isinstance(pd, dict):
        # fix types when just a value is passed (tests)
        pytype = guess_type(pd)
        if pytype not in stypes:
            pd = serialize_data(pd)
        return {"value":pd, "type": pytype}
    pydata = pd["value"]
    pytype = pd["type"].lower()
    logger.debug("pydata value provided: '%s' with type '%s', %s",
                 pydata, type(pydata), pd["type"])
    empty_strings = ["None", ""]
    if pydata in empty_strings:
        if pytype in ["string", "str"]:
            pydata = ''
        else:
            pydata = bytes() # Treat None/Empty objects as empty object data.
            pytype = 'object'
    builtin_types = get_builtins()
    if pytype != "raw" and type(pydata) in builtin_types.values() and pytype not in builtin_types.keys():
        logger.warning("Type of pydata %s provided differs from specified type: %s", type(pydata).__name__, pytype)
        # pytype = type(pydata).__name__
    if pytype == "json":
        try:
            pydata = json.loads(pydata)
        except JSONDecodeError:
            pydata = pydata.encode()
    if pytype == "eval":
        # try:
        pydata = eval(pydata) # pylint: disable=eval-used
        # except Exception:
        #     pydata = pydata.encode()
    elif pytype in ["int", "integer"]:
        try:
            pydata = int(pydata)
            pytype = "int"
        except ValueError:
            pydata = pydata.encode()
    elif pytype == "float":
        try:
            pydata = float(pydata)
        except ValueError:
            pydata = pydata.encode()
    elif pytype in ["boolean","bool"]:
        try:
            pydata = bool(pydata)
            pytype = "bool"
        except ValueError:
            pydata = pydata.encode()
    elif pytype.split(".")[0].lower() == "object":
        if pydata:
            try:
                pydata = base64.b64decode(pydata.encode())
                pydata = dill.loads(pydata)
            except (TypeError, dill.PickleError) as e:
                logger.error("Issue loading pydata object with pickle %s", pydata)
                raise dill.PickleError from e
    elif pytype.lower() == "raw":
        pydata = dill.loads(base64.b64decode(pydata))
        logger.debug("Returning pydata of type: %s", type(pydata))
        # return pydata
    pd["value"] = dill.dumps(pydata)
    pd["type"] = pytype
    return pd


##
# @brief Memory
# @details In-memory storage of intermediate data products
# @par EAGLE_START
# @param category Memory
# @param tag daliuge
# @param pydata None/Object/ApplicationArgument/NoPort/ReadWrite//False/False/Data to be loaded into memory
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param dropclass dlg.data.drops.memory.InMemoryDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/Drop class
# @param base_name memory/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @par EAGLE_END
class InMemoryDROP(DataDROP):
    """
    A DROP that points data stored in memory.
    """

    # Allow in-memory drops to be automatically removed by default

    @track_current_drop
    def __init__(self, *args, **kwargs):
        if "persist" not in kwargs:
            kwargs["persist"] = False
        if "expireAfterUse" not in kwargs:
            kwargs["expireAfterUse"] = True
        if kwargs["persist"]:
            kwargs["expireAfterUse"] = False
        super().__init__(*args, **kwargs)

    def initialize(self, **kwargs):
        """
        If there is a pydata argument use that to populate the DROP

        Note: It is essential to have the type of the initial PyData correct, or it could
        cause issues with the type of IO system that is used.

        If parse_pydata() determines that the data passed via 'pydata' is a String, it
        will use StringIO as the buffer. This can have ramifications for none-string
        encodings used further down the track.
        """
        # TODO: We need to ignore pydata if we have a connected input port
        args = []
        pdict = {"value":None, "type": "raw"}  # initialize this value to enforce BytesIO
        field_names = (
            [f["name"] for f in kwargs["fields"]] if "fields" in kwargs else []
        )
        if "pydata" in kwargs and not (
            "fields" in kwargs and "pydata" in field_names
        ):  # means that is was passed directly (e.g. from tests)
            pydata = kwargs.pop("pydata")
            pdict = parse_pydata(pydata)
        elif "fields" in kwargs and "pydata" in field_names:
            data_pos = field_names.index("pydata")
            pdict = kwargs["fields"][data_pos].copy()
            pdict = parse_pydata(pdict)
        self.data_type = pdict["type"]
        if pdict and pdict["type"] in ["str","string"]:
            self.data_type =  "String" if pdict["value"] else "raw"
        else:
            self.data_type = pdict["type"] if pdict else ""
        if pdict["value"]:
            args.append(pdict["value"])
            logger.debug("Loaded into memory: %s, %s, %s", pdict["value"], self.data_type, type(pdict["value"]))
        self._buf = io.BytesIO(*args) if self.data_type != "String" else io.StringIO(
            dill.loads(*args)) # pylint: disable = no-value-for-parameter
        self.size = len(pdict["value"]) if pdict["value"] else 0

    def getIO(self):
        if (
            hasattr(self, "_tp")
            and hasattr(self, "_sessionId")
            and sys.version_info >= (3, 8)
        ):
            return SharedMemoryIO(self.oid, self._sessionId)
        else:
            return MemoryIO(self._buf)

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]
        return "mem://%s/%d/%d" % (hostname, os.getpid(), id(self._buf))

    # Override
    def generate_reproduce_data(self):
        from dlg.droputils import allDropContents

        data = b""
        try:
            data = allDropContents(self, self.size)
        except IOError:
            logger.debug("Could not read drop reproduce data")
        return {"data_hash": common_hash(data)}


##
# @brief PythonObject
# @details A Python object stored in memory
# @par EAGLE_START
# @param category PythonObject
# @param tag daliuge
# @param self /Object/ComponentParameter/InputOutput/ReadWrite//False/False/Reference to object
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Object should be serialized
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in the object
# @param dropclass dlg.data.drops.memory.InMemoryDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/Drop class
# @param base_name memory/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of class
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @par EAGLE_END
class PythonObjectDROP(InMemoryDROP):
    """
    A python object is an InMemoryDROP but has a special category of PythonObject.

    This is only here to force the creation of the component for the palette.
    """


##
# @brief SharedMemory
# @details Data stored in shared memory
# @par EAGLE_START
# @param category SharedMemory
# @param tag daliuge
# @param pydata None/String/ApplicationArgument/NoPort/ReadWrite//False/False/Data to be loaded into memory
# @param io /Object/ApplicationArgument/InputOutput/ReadWrite//False/False/Input Output port
# @param persist False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component contains data that should not be deleted after execution
# @param block_skip False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/If set the drop will block a skipping chain until the last producer has finished and is not also skipped.
# @param data_volume 5/Float/ConstraintParameter/NoPort/ReadWrite//False/False/Estimated size of the data contained in this node
# @param group_end False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Is this node the end of a group?
# @param streaming False/Boolean/ComponentParameter/NoPort/ReadWrite//False/False/Specifies whether this data component streams input and output data
# @param dropclass dlg.data.drops.memory.SharedMemoryDROP/String/ComponentParameter/NoPort/ReadOnly//False/False/Drop class
# @param base_name memory/String/ComponentParameter/NoPort/ReadOnly//False/False/Base name of application class
# @par EAGLE_END
class SharedMemoryDROP(DataDROP):
    """
    A DROP that points to data stored in shared memory.
    This drop is functionality equivalent to an InMemory drop running in a concurrent environment.
    In this case however, the requirement for shared memory is explicit.

    @WARNING Currently implemented as writing to shmem and there is no backup behaviour.
    """

    def initialize(self, **kwargs):
        args = []
        pydata = None
        if "pydata" in kwargs and not (
            "nodeAttributes" in kwargs and "pydata" in kwargs["nodeAttributes"]
        ):  # means that is was passed directly
            pydata = kwargs.pop("pydata")
            logger.debug("pydata value provided: %s", pydata)
            try:  # test whether given value is valid
                _ = dill.loads(base64.b64decode(pydata.encode("latin1")))
                pydata = base64.b64decode(pydata.encode("latin1"))
            except PickleError:
                pydata = None
                logger.warning("Unable to load using pickle, default to None")
            except binascii.Error:
                pydata = None
                logger.warning("Unable to load using base64, defaulting to None")
        elif "nodeAttributes" in kwargs and "pydata" in kwargs["nodeAttributes"]:
            pydata = parse_pydata(kwargs["nodeAttributes"]["pydata"])
        args.append(pydata)
        self._buf = io.BytesIO(*args)

    def getIO(self):
        if sys.version_info >= (3, 8):
            if hasattr(self, "_sessionId"):
                return SharedMemoryIO(self.oid, self._sessionId)
            else:
                # Using Drop without manager, just generate a random name.
                sess_id = "".join(
                    random.choices(string.ascii_uppercase + string.digits, k=10)
                )
                return SharedMemoryIO(self.oid, sess_id)
        else:
            raise NotImplementedError(
                "Shared memory is only available with Python >= 3.8"
            )

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]
        return f"shmem://{hostname}/{os.getpid()}/{id(self._buf)}"



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
This package contains several general-purpose applications in form of
DROPs that we have developed as examples and for real-life use. Most of them
are based on the :class:`BarrierAppDROP`.
"""


def get_include_dir():
    """The path that C/C++ applications need to add to use the dlg.h header"""
    import os

    return os.path.dirname(__file__)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing the REST layer that exposes the methods of the different
Data Managers (DROPManager and DataIslandManager) to the outside world.
"""
# pylint: disable=protected-access

from email.message import Message
import functools
import io
import json
import logging
import os
import re
import tarfile
import threading

import bottle

from bottle import static_file
from pathlib import Path

from dlg import constants
from dlg.manager.client import NodeManagerClient, DataIslandManagerClient
from dlg import utils
from dlg.exceptions import (
    InvalidGraphException,
    InvalidSessionState,
    DaliugeException,
    NoSessionException,
    SessionAlreadyExistsException,
    InvalidDropException,
    InvalidRelationshipException,
    SubManagerException,
)
from dlg.restserver import RestServer
from dlg.restutils import RestClient, RestClientException
from dlg.manager.session import generateLogFileName
from dlg.common.deployment_methods import DeploymentMethods
from dlg.manager.manager_data import Node

logger = logging.getLogger(f"dlg.{__name__}")

def file_as_string(fname, enc="utf8"):
    res = Path(__file__).parent / fname
    return utils.b2s(res.read_bytes(), enc)


def daliuge_aware(func):
    @functools.wraps(func)
    def fwrapper(*args, **kwargs):
        try:
            res = func(*args, **kwargs)

            if isinstance(res, (bytes, bottle.HTTPResponse)):
                return res

            if res is not None:
                # set CORS headers
                origin = bottle.request.headers.raw("Origin")
                # logger.debug("CORS request comming from: %s", origin)
                # logger.debug("Request method: %s", bottle.request.method)
                if origin is None or re.match(
                    r"(http://dlg-trans.local:80[0-9][0-9]|https://dlg-trans.icrar.org)",
                    origin,
                ):
                    pass
                elif re.match(r"http://((localhost)|(127.0.0.1)):80[0-9][0-9]", origin):
                    origin = "http://localhost:8084"

                bottle.response.headers["Access-Control-Allow-Origin"] = origin
                bottle.response.headers["Access-Control-Allow-Credentials"] = "true"
                bottle.response.headers["Access-Control-Allow-Methods"] = (
                    "GET, POST, PUT, OPTIONS, HEAD"
                )
                bottle.response.headers["Access-Control-Allow-Headers"] = (
                    "Origin, Accept, Content-Type, Content-Encoding, X-Requested-With, X-CSRF-Token"
                )
                # logger.debug("CORS headers set to allow from: %s", origin)
            bottle.response.content_type = "application/json"
            # logger.debug("REST function called: %s", func.__name__)
            jres = (
                json.dumps(res)
                if res is not None
                else json.dumps({"Status": "Success"})
            )
            # logger.debug("Bottle sending back result: %s", jres[: min(len(jres), 80)])
            return jres
        except Exception as e: # pylint: disable=broad-exception-caught
            logger.exception("Error while fulfilling request for func %s ", func)

            status, eargs = 500, ()
            if isinstance(e, NotImplementedError):
                status, eargs = 501, e.args
            elif isinstance(e, NoSessionException):
                status, eargs = 404, (e.session_id,)
            elif isinstance(e, SessionAlreadyExistsException):
                status, eargs = 409, (e.session_id,)
            elif isinstance(e, InvalidDropException):
                status, eargs = 409, ((e.oid, e.uid), e.reason)
            elif isinstance(e, InvalidRelationshipException):
                status, eargs = 409, (e.rel, e.reason)
            elif isinstance(e, InvalidGraphException):
                status, eargs = 400, e.args
            elif isinstance(e, InvalidSessionState):
                status, eargs = 400, e.args
            elif isinstance(e, RestClientException):
                status, eargs = 556, e.args
            elif isinstance(e, SubManagerException):
                status = 555
                eargs = {}
                # args[1] is a dictionary of host:exception
                for host, subex in e.args[1].items():
                    logger.debug(">>>> Error class name: %s", subex.__class__.__name__)
                    eargs[host] = {
                        "type": subex.__class__.__name__,
                        # "args": subex.args,
                        "args": "dummy",
                    }
            elif isinstance(e, DaliugeException):
                status, eargs = 555, e.args
            else:
                raise

            error = {"type": e.__class__.__name__, "args": eargs}
            bottle.response.status = status
            return json.dumps(error)

    return fwrapper


class ManagerRestServer(RestServer):
    """
    An object that wraps a DataManager and exposes its methods via a REST
    interface. The server is started via the `start` method in a separate thread
    and runs until the process is shut down.

    This REST server currently also serves HTML pages in some of its methods
    (i.e. those not under /api).
    """

    def __init__(self, dm, maxreqsize=10):
        super(ManagerRestServer, self).__init__()

        # Increase maximum file sizes
        bottle.BaseRequest.MEMFILE_MAX = maxreqsize * 1024 * 1024

        self.dm = dm

        # Mappings
        app = self.app
        app.get("/api/submission_method", callback=self.submit_methods)
        app.post("/api/stop", callback=self.stop_manager)
        app.post("/api/sessions", callback=self.createSession)
        app.get("/api/sessions", callback=self.getSessions)
        app.get("/api/sessions/<sessionId>", callback=self.getSessionInformation)
        app.delete("/api/sessions/<sessionId>", callback=self.destroySession)
        app.get("/api/sessions/<sessionId>/logs", callback=self.getLogFile)
        app.get("/api/sessions/<sessionId>/status", callback=self.getSessionStatus)
        app.post("/api/sessions/<sessionId>/deploy", callback=self.deploySession)
        app.post("/api/sessions/<sessionId>/cancel", callback=self.cancelSession)
        app.get("/api/sessions/<sessionId>/graph", callback=self.getGraph)
        app.get("/api/sessions/<sessionId>/graph/size", callback=self.getGraphSize)
        app.get(
            "/api/sessions/<sessionId>/graph/status",
            callback=self.getGraphStatus,
        )
        app.post(
            "/api/sessions/<sessionId>/graph/append",
            callback=self.addGraphSpec,
        )
        app.get(
            "/api/sessions/<sessionId>/repro/data",
            callback=self.getSessionReproData,
        )
        app.get(
            "/api/sessions/<sessionId>/repro/status",
            callback=self.getSessionReproStatus,
        )

        app.route("/api/sessions", method="OPTIONS", callback=self.acceptPreflight)
        app.route(
            "/api/sessions/<sessionId>/graph/append",
            method="OPTIONS",
            callback=self.acceptPreflight2,
        )

        # The non-REST mappings that serve HTML-related content
        app.route("/static/<filepath:path>", callback=self.server_static)
        app.get("/session", callback=self.visualizeSession)
        app.route("/api/sessions/<sessionId>/dir", callback=self._getSessionDir)
        app.route("/api/sessions/<sessionId>/graph/drop/<dropId>",
                  callback=self._getDropStatus)
        app.route("/sessions/<sessionId>/graph/drop/<dropId>",
                callback=self.getDropStatus)

        # sub-class specifics
        self.initializeSpecifics(app)

    def initializeSpecifics(self, app):
        """
        Methods through which subclasses can initialize other mappings on top of
        the default ones and perform other DataManager-specific actions.
        The default implementation does nothing.
        """

    @daliuge_aware
    def submit_methods(self):
        return {"methods": [DeploymentMethods.BROWSER, DeploymentMethods.SERVER]}

    def _stop_manager(self):
        self.dm.shutdown()
        self.stop()
        logger.info(
            "Thanks for using our %s, come back again :-)",
            self.dm.__class__.__name__,
        )

    @daliuge_aware
    def stop_manager(self):
        threading.Thread(target=self._stop_manager).start()

    @daliuge_aware
    def createSession(self):
        newSession = bottle.request.json
        sessionId = newSession["sessionId"]
        self.dm.createSession(sessionId)
        return {"sessionId": sessionId}

    @daliuge_aware
    def acceptPreflight(self):
        return {}

    @daliuge_aware
    def acceptPreflight2(self, sessionId):
        logger.info("Preflight2 for %s", sessionId)
        return {}

    def sessions(self):
        sessions = []
        for sessionId in self.dm.getSessionIds():
            sessions.append(
                {
                    "sessionId": sessionId,
                    "status": self.dm.getSessionStatus(sessionId),
                    "size": self.dm.getGraphSize(sessionId),
                }
            )
        return sessions

    @daliuge_aware
    def getSessions(self):
        return self.sessions()

    @daliuge_aware
    def getSessionInformation(self, sessionId):
        status = self.dm.getSessionStatus(sessionId)
        try:
            graphDict = self.dm.getGraph(sessionId)
            directory = self.dm.getSessionDir(sessionId)
        except KeyError:  # Pristine state sessions don't have a graph, yet.
            graphDict = {}
            status = 0
            directory = ""
        return {"status": status, "graph": graphDict, "dir":directory}

    @daliuge_aware
    def getSessionReproStatus(self, sessionId):
        return self.dm.getSessionReproStatus(sessionId)

    @daliuge_aware
    def getSessionsReproStatus(self):
        sessions = []
        for sessionId in self.dm.getSessionIds():
            sessions.append(
                {
                    "sessionId": sessionId,
                    "status": self.dm.getSessionStatus(sessionId),
                    "size": self.dm.getGraphSize(sessionId),
                    "repro": self.dm.getSessionReproStatus(sessionId),
                }
            )
        return sessions

    @daliuge_aware
    def getSessionReproData(self, sessionId):
        #  For now, we only have information on a per-graph basis.
        graphDict = self.dm.getGraph(sessionId)
        reprodata = self.dm.getGraphReproData(sessionId)
        return {"graph": graphDict, "reprodata": reprodata}

    @daliuge_aware
    def destroySession(self, sessionId):
        self.dm.destroySession(sessionId)

    @daliuge_aware
    def getSessionStatus(self, sessionId):
        return self.dm.getSessionStatus(sessionId)

    @daliuge_aware
    def deploySession(self, sessionId):
        completedDrops = []
        if "completed" in bottle.request.forms:
            completedDrops = bottle.request.forms["completed"].split(",")
        return self.dm.deploySession(sessionId, completedDrops=completedDrops)
        # return {"Status": "Success"}

    @daliuge_aware
    def cancelSession(self, sessionId):
        self.dm.cancelSession(sessionId)

    @daliuge_aware
    def getGraph(self, sessionId):
        return self.dm.getGraph(sessionId)

    @daliuge_aware
    def getGraphSize(self, sessionId):
        return self.dm.getGraphSize(sessionId)

    @daliuge_aware
    def getGraphStatus(self, sessionId):
        return self.dm.getGraphStatus(sessionId)

    @daliuge_aware
    def addGraphSpec(self, sessionId):
        # WARNING: TODO: Somehow, the content_type can be overwritten to 'text/plain'
        logger.debug("Graph content type: %s", bottle.request.content_type)
        if (
            "application/json" not in bottle.request.content_type
            and "text/plain" not in bottle.request.content_type
        ):
            bottle.response.status = 415
            return

        # We also accept gzipped content
        hdrs = bottle.request.headers
        logger.debug("Graph hdr: %s", {k: v for k, v in hdrs.items()})
        if hdrs.get("Content-Encoding", None) == "gzip":
            json_content = utils.ZlibUncompressedStream(bottle.request.body)
        else:
            json_content = bottle.request.body

        graph_parts = bottle.json_loads(json_content.read())

        # Do something about host Nodes in graph_parts?
        return self.dm.addGraphSpec(sessionId, graph_parts)
        # return {"graph_parts": graph_parts}

    # ===========================================================================
    # non-REST methods
    # ===========================================================================
    def server_static(self, filepath):
        staticRoot = Path(__file__).parent / "web/static"
        return bottle.static_file(filepath, root=staticRoot)

    def _getSessionDir(self, sessionId):
        return self.dm.getSessionDir(sessionId)

    def visualizeSession(self):
        params = bottle.request.params
        sessionId = params["sessionId"] if "sessionId" in params else ""
        selectedNode = params["node"] if "node" in params else ""
        viewMode = params["view"] if "view" in params else ""
        tpl = file_as_string("web/session.html")
        urlparts = bottle.request.urlparts
        serverUrl = urlparts.scheme + "://" + urlparts.netloc
        return bottle.template(
            tpl,
            sessionId=sessionId,
            selectedNode=selectedNode,
            viewMode=viewMode,
            serverUrl=serverUrl,
            dmType=self.dm.__class__.__name__,
            sessionDir=sessionId
        )

    def _getDropStatus(self, sessionId, dropId):
        return self.dm.getDropStatus(sessionId, dropId)

    def getDropStatus(self, sessionId, dropId):
        params = bottle.request.params
        logger.warning("PARAMS: %s", params)

        urlparts = bottle.request.urlparts
        serverUrl = urlparts.scheme + "://" + urlparts.netloc

        data = self._getDropStatus(sessionId, dropId)
        if data["logs"]:
            columns = [col for col in data["logs"][-1].keys()]
            filter_column = "Level"
            filter_column_index = columns.index(filter_column)
        else:
            columns = []
            filter_column_index=0

        tpl = file_as_string("web/drop_log.html")
        return bottle.template(
            tpl,
            data=data,
            columns=columns,
            filter_index=filter_column_index,
            sessionId=sessionId,
            serverUrl=serverUrl,
            dmType=self.dm.__class__.__name__,
        )

class NMRestServer(ManagerRestServer):
    """
    A REST server for NodeManagers. It includes mappings for NM-specific
    methods and the mapping for the main visualization HTML pages.
    """

    def initializeSpecifics(self, app):
        app.get("/api", callback=self.getNMStatus)
        app.post(
            "/api/sessions/<sessionId>/graph/link",
            callback=self.linkGraphParts,
        )
        app.post(
            "/api/sessions/<sessionId>/subscriptions",
            callback=self.add_node_subscriptions,
        )
        app.post("/api/sessions/<sessionId>/trigger", callback=self.trigger_drops)
        # The non-REST mappings that serve HTML-related content
        app.get("/", callback=self.visualizeDM)
        app.get("/api/shutdown", callback=self.shutdown_node_manager)

    @daliuge_aware
    def shutdown_node_manager(self):
        logger.debug("Shutting down node manager")
        self.dm.shutdown()

    @daliuge_aware
    def getNMStatus(self):
        # we currently return the sessionIds, more things might be added in the
        # future
        logger.debug("NM REST call: status")
        return {"sessions": self.sessions()}

    @daliuge_aware
    def getLogFile(self, sessionId):
        logger.debug("NM REST call: logfile")
        logdir = self.dm.getLogDir()
        logfile = generateLogFileName(logdir, sessionId)
        if not os.path.isfile(logfile):
            raise NoSessionException(sessionId, "Log file not found.")
        return static_file(
            os.path.basename(logfile),
            root=logdir,
            download=os.path.basename(logfile),
        )

    @daliuge_aware
    def linkGraphParts(self, sessionId):
        logger.debug("NM REST call: graph/link")
        params = bottle.request.params
        lhOID = params["lhOID"]
        rhOID = params["rhOID"]
        linkType = int(params["linkType"])
        self.dm.linkGraphParts(sessionId, lhOID, rhOID, linkType)

    @daliuge_aware
    def add_node_subscriptions(self, sessionId):
        logger.debug("NM REST call: add_subscriptions %s", bottle.request.json)
        if bottle.request.content_type != "application/json":
            bottle.response.status = 415
            return
        subscriptions = self._parse_subscriptions(bottle.request.json)
        self.dm.add_node_subscriptions(sessionId, subscriptions)

    def _parse_subscriptions(self, json_request):
        return {Node(host): droprels for host, droprels in json_request.items()}

    @daliuge_aware
    def trigger_drops(self, sessionId):
        if bottle.request.content_type != "application/json":
            bottle.response.status = 415
            return
        self.dm.trigger_drops(sessionId, bottle.request.json)

    # ===========================================================================
    # non-REST methods
    # ===========================================================================
    def visualizeDM(self):
        tpl = file_as_string("web/dm.html")
        urlparts = bottle.request.urlparts
        serverUrl = urlparts.scheme + "://" + urlparts.netloc
        return bottle.template(
            tpl,
            serverUrl=serverUrl,
            dmType=self.dm.__class__.__name__,
            reset="false",
        )


class CompositeManagerRestServer(ManagerRestServer):
    """
    A REST server for DataIslandManagers. It includes mappings for DIM-specific
    methods.
    """

    def initializeSpecifics(self, app):
        app.get("/api", callback=self.getCMStatus)
        app.get("/api/past_sessions", callback=self.getPastSessions)
        app.get("/api/nodes", callback=self.getCMNodes)
        app.post("/api/node/<node>", callback=self.addCMNode)
        app.delete("/api/node/<node>", callback=self.removeCMNode)

        # Query forwarding to sub-nodes
        app.get("/api/node/<node>/sessions", callback=self.getNodeSessions)
        app.get(
            "/api/node/<node>/sessions/<sessionId>",
            callback=self.getNodeSessionInformation,
        )
        app.get(
            "/api/node/<node>/sessions/<sessionId>/status",
            callback=self.getNodeSessionStatus,
        )
        app.get(
            "/api/node/<node>/sessions/<sessionId>/graph",
            callback=self.getNodeGraph,
        )
        app.get(
            "/api/node/<node>/sessions/<sessionId>/graph/status",
            callback=self.getNodeGraphStatus,
        )

        # The non-REST mappings that serve HTML-related content
        app.get("/", callback=self.visualizeDIM)

    @daliuge_aware
    def getCMStatus(self):
        """
        REST (GET): /api/

        Return JSON-compatible list of Composite Manager nodes and sessions
        """
        return {
            "hosts": [str(n) for n in self.dm.dmHosts],
            "sessionIds": self.dm.getSessionIds(),
        }

    @daliuge_aware
    def getCMNodes(self):
        """
        REST (GET): /api/nodes

        Return JSON-compatible list of Composite Manager nodes
        """
        return [str(n) for n in self.dm.nodes]

    def _getAllCMNodes(self):
        return self.dm.nodes

    @daliuge_aware
    def addCMNode(self, node):
        """
        REST (POST): "/api/node/<node>"

        Add the posted node to the Composite Manager

        Converts from JSON to our ser
        """
        logger.debug("Adding node %s", node)
        self.dm.add_node(Node(node))

    @daliuge_aware
    def removeCMNode(self, node):
        """
        REST (DELETE): "/api/node/<node>"

        Add the posted node to the Composite Manager

        """
        logger.debug("Removing node %s", node)
        self.dm.remove_node(Node(node))

    @daliuge_aware
    def getNodeSessions(self, node):
        """
        REST (GET): "/api/node/<node>/sessions"

        Retrieve sessions for given node
        """
        host_node = Node(node)
        if host_node not in self.dm.nodes:
            raise RuntimeError(f"{host_node} not in current list of nodes")
        with NodeManagerClient(host=host_node.host, port=host_node.port) as dm:
            return dm.sessions()

    @daliuge_aware
    def getPastSessions(self):
        """
        REST (GET): /api/past_sessions

        Return JSON-compatible list of Composite Manager nodes
        """
        return self.pastSessions()

    def pastSessions(self):
        """
        Retrieve sessions from this DropManager and place it in JSON-format, for
        serialisation across the wire.
        """

        return [
            {"sessionId": pastSession} for pastSession in self.dm.getPastSessionIds()
        ]

    def _tarfile_write(self, tar, headers, stream):
        file_header = headers.getheader("Content-Disposition")
        length = headers.getheader("Content-Length")
        m = Message()
        m.add_header("content-disposition", file_header)
        filename = m.get_params("filename")
        info = tarfile.TarInfo(filename)
        info.size = int(length)

        content = []
        while True:
            buffer = stream.read()
            if not buffer:
                break
            content.append(buffer)

        tar.addfile(info, io.BytesIO(initial_bytes="".join(content).encode()))

    @daliuge_aware
    def getLogFile(self, sessionId):
        fh = io.BytesIO()
        with tarfile.open(fileobj=fh, mode="w:gz") as tar:
            for node in self._getAllCMNodes():
                with NodeManagerClient(host=node.host, port=node.port) as dm:
                    try:
                        stream, resp = dm.get_log_file(sessionId)
                        self._tarfile_write(tar, resp, stream)
                    except NoSessionException:
                        pass

        data = fh.getvalue()
        size = len(data)
        bottle.response.set_header("Content-type", "application/x-tar")
        bottle.response["Content-Disposition"] = (
            f"attachment; " f"filename=dlg_{sessionId}.tar"
        )
        bottle.response["Content-Length"] = size
        return data

    @daliuge_aware
    def getNodeSessionInformation(self, node: str, sessionId):
        try:
            n = Node(self.dm.get_node_from_json(node))
            with NodeManagerClient(host=n.host, port=n.port) as dm:
                return dm.session(sessionId)
        except ValueError as e:
            raise ValueError(f"{n} not in current list of nodes") from e

    @daliuge_aware
    def getNodeSessionStatus(self, node: str, sessionId):
        try:
            n = Node(self.dm.get_node_from_json(node))
            with NodeManagerClient(host=n, port=n.port) as dm:
                return dm.session_status(sessionId)
        except ValueError as e:
            raise ValueError(f"{node} not in current list of nodes") from e

    @daliuge_aware
    def getNodeGraph(self, node: str, sessionId):
        try:
            n = Node(self.dm.get_node_from_json(node))
            with NodeManagerClient(host=n.host, port=n.port) as dm:
                return dm.graph(sessionId)
        except ValueError as e:
            raise ValueError(f"{node} not in current list of nodes") from e

    @daliuge_aware
    def getNodeGraphStatus(self, node: str, sessionId):
        try:
            n = Node(self.dm.get_node_from_json(node))
            with NodeManagerClient(host=n.host, port=n.port) as dm:
                return dm.graph_status(sessionId)
        except ValueError as e:
            raise ValueError(f"{node} not in current list of nodes") from e

    # ===========================================================================
    # non-REST methods
    # ===========================================================================

    def visualizeDIM(self):
        tpl = file_as_string("web/dim.html")
        urlparts = bottle.request.urlparts
        selectedNode = (
            bottle.request.params["node"] if "node" in bottle.request.params else ""
        )
        serverUrl = urlparts.scheme + "://" + urlparts.netloc
        return bottle.template(
            tpl,
            dmType=self.dm.__class__.__name__,
            dmPort=self.dm.dmPort,
            serverUrl=serverUrl,
            dmHosts=json.dumps([str(n) for n in self.dm.dmHosts]),
            nodes=json.dumps([str(n) for n in self.dm.nodes]),
            selectedNode=selectedNode,
        )


class MasterManagerRestServer(CompositeManagerRestServer):
    def initializeSpecifics(self, app):
        CompositeManagerRestServer.initializeSpecifics(self, app)
        # DIM manamagement
        app.post("/api/island/<dim>", callback=self.addDIM)
        app.delete("/api/island/<dim>", callback=self.removeDIM)
        # Query forwarding to daemons
        app.post("/api/managers/<host>/island", callback=self.createDataIsland)
        app.post("/api/managers/<host>/node/start", callback=self.startNM)
        app.post("/api/managers/<host>/node/stop", callback=self.stopNM)
        # manage node manager assignment
        app.post("/api/managers/<host>/node/<node>", callback=self.addNM)
        app.delete("/api/managers/<host>/node/<node>", callback=self.removeNM)
        # Querying about managers
        app.get("/api/islands", callback=self.getDIMs)
        app.get("/api/nodes", callback=self.getNMs)
        app.get("/api/managers/<host>/node", callback=self.getNMInfo)
        app.get("/api/managers/<host>/island", callback=self.getDIMInfo)
        app.get("/api/managers/<host>/master", callback=self.getMMInfo)

    @daliuge_aware
    def createDataIsland(self, host):
        with RestClient(
            host=host, port=constants.DAEMON_DEFAULT_REST_PORT, timeout=10
        ) as c:
            c._post_json("/managers/island/start", bottle.request.body.read())
        self.dm.addDmHost(host)
        return {"islands": self.dm.dmHosts}

    @daliuge_aware
    def getDIMs(self):
        return {"islands": self.dm.dmHosts}

    @daliuge_aware
    def addDIM(self, dim):
        logger.debug("Adding DIM %s", dim)
        self.dm.addDmHost(dim)

    @daliuge_aware
    def removeDIM(self, dim):
        logger.debug("Removing dim %s", dim)
        self.dm.removeDmHost(dim)

    @daliuge_aware
    def getNMs(self):
        return {"nodes": [str(n) for n in self.dm.nodes]}

    @daliuge_aware
    def startNM(self, host):
        port = constants.DAEMON_DEFAULT_REST_PORT
        logger.debug("Sending NM start request to %s:%s", host, port)
        with RestClient(host=host, port=port, timeout=10) as c:
            return json.loads(c.POST("/managers/node/start").read())

    @daliuge_aware
    def stopNM(self, host):
        port = constants.DAEMON_DEFAULT_REST_PORT
        logger.debug("Sending NM stop request to %s:%s", host, port)
        with RestClient(host=host, port=port, timeout=10) as c:
            return json.loads(c.POST("/managers/node/stop").read())

    @daliuge_aware
    def addNM(self, host, node):
        port = constants.ISLAND_DEFAULT_REST_PORT
        logger.debug("Adding NM %s to DIM %s", node, host)
        with RestClient(host=host, port=port, timeout=10, url_prefix="/api") as c:
            return json.loads(
                c.POST(
                    f"/node/{node}",
                ).read()
            )

    @daliuge_aware
    def removeNM(self, host, node):
        port = constants.ISLAND_DEFAULT_REST_PORT
        logger.debug("Removing NM %s from DIM %s", node, host)
        with RestClient(host=host, port=port, timeout=10, url_prefix="/api") as c:
            return json.loads(c._DELETE(f"/node/{node}").read())

    @daliuge_aware
    def getNMInfo(self, host):
        port = constants.DAEMON_DEFAULT_REST_PORT
        logger.debug("Sending request %s:%s/managers/node", host, port)
        with RestClient(host=host, port=port, timeout=10) as c:
            return json.loads(c._GET("/managers/node").read())

    @daliuge_aware
    def getDIMInfo(self, host):
        with RestClient(
            host=host, port=constants.DAEMON_DEFAULT_REST_PORT, timeout=10
        ) as c:
            return json.loads(c._GET("/managers/island").read())

    @daliuge_aware
    def getMMInfo(self, host):
        with RestClient(
            host=host, port=constants.DAEMON_DEFAULT_REST_PORT, timeout=10
        ) as c:
            return json.loads(c._GET("/managers/master").read())

    def _getAllCMNodes(self):
        nodes = []
        for host in self.dm.dmHosts:
            h = Node(host)
            with DataIslandManagerClient(host=h.host, port=h.port) as dm:
                nodes += dm.nodes()
        return [str(n) for n in nodes]



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import unittest
import json
import daliuge_tests.engine.graphs as test_graphs

from importlib.resources import files

from dlg import graph_loader
from dlg.ddap_protocol import DROPLinkType, DROPRel
from dlg.data.drops.container import ContainerDROP
from dlg.apps.app_base import AppDROP

from dlg.data.drops.memory import InMemoryDROP, SharedMemoryDROP
from dlg.data.drops.directory import DirectoryDROP
from dlg.apps.simple import RandomArrayApp


# Used in the textual representation of the graphs in these tests
class DummyApp(AppDROP):
    pass


class TestGraphLoader(unittest.TestCase):
    def test_singleMemoryDrop(self):
        dropSpecList = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            }
        ]
        a = graph_loader.createGraphFromDropSpecList(dropSpecList)[0]
        self.assertIsInstance(a, InMemoryDROP)
        self.assertEqual("A", a.oid)
        self.assertEqual("A", a.uid)

    def test_sharedMemoryDrop(self):
        dropSpecList = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.SharedMemoryDROP",
            }
        ]
        a = graph_loader.createGraphFromDropSpecList(dropSpecList)[0]
        self.assertIsInstance(a, SharedMemoryDROP)
        self.assertEqual("A", a.oid)
        self.assertEqual("A", a.uid)

    def test_containerDrop(self):
        dropSpecList = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
            },
            {
                "oid": "B",
                "categoryType": "container",
                "dropclass": "dlg.data.drops.container.ContainerDROP",
                "children": ["A"],
            },
        ]
        a = graph_loader.createGraphFromDropSpecList(dropSpecList)[0]
        self.assertIsInstance(a, InMemoryDROP)
        self.assertEqual("A", a.oid)
        self.assertEqual("A", a.uid)
        self.assertIsNotNone(a.parent)
        b = a.parent
        self.assertIsInstance(b, ContainerDROP)
        self.assertEqual("B", b.oid)
        self.assertEqual("B", b.uid)

    def test_directoryDROP(self):

        # A directory container
        dropSpecList = [
            {
                "oid": "B",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.DirectoryDROP",
                "dirname": "/tmp/testdir"
            },
        ]
        b = graph_loader.createGraphFromDropSpecList(dropSpecList)[0]
        self.assertIsInstance(b, DirectoryDROP)
        self.assertEqual("/tmp/testdir", b.path)

    def test_consumer(self):
        dropSpecList = [
            {
                "oid": "A",
                "categoryType": "Data",
                "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                "consumers": ["B"],
            },
            {
                "oid": "B",
                "categoryType": "Application",
                "dropclass": "test.test_graph_loader.DummyApp",
            },
        ]
        a = graph_loader.createGraphFromDropSpecList(dropSpecList)[0]
        self.assertIsInstance(a, InMemoryDROP)
        self.assertEqual("A", a.oid)
        self.assertEqual("A", a.uid)
        self.assertEqual(1, len(a.consumers))
        b = a.consumers[0]
        self.assertIsInstance(b, DummyApp)
        self.assertEqual("B", b.oid)
        self.assertEqual("B", b.uid)
        self.assertEqual(a, b.inputs[0])

    def test_removeUnmetRelationships(self):
        # Unmet relationsips are
        # DROPRel(D, CONSUMER, A)
        # DROPRel(D, STREAMING_CONSUMER, C)
        # DROPRel(Z, PRODUCER, A)
        # DROPRel(X, PRODUCER, A)
        graphDesc = [
            {"oid": "A", "consumers": ["B", "D"], "producers": ["Z", "X"]},
            {"oid": "B", "outputs": ["C"]},
            {"oid": "C", "streamingConsumers": ["D"]},
        ]

        unmetRelationships = graph_loader.removeUnmetRelationships(graphDesc)
        self.assertEqual(4, len(unmetRelationships))
        self.assertIn(
            DROPRel("D", DROPLinkType.CONSUMER, "A"), unmetRelationships
        )
        self.assertIn(
            DROPRel("D", DROPLinkType.STREAMING_CONSUMER, "C"),
            unmetRelationships,
        )
        self.assertIn(
            DROPRel("Z", DROPLinkType.PRODUCER, "A"), unmetRelationships
        )
        self.assertIn(
            DROPRel("X", DROPLinkType.PRODUCER, "A"), unmetRelationships
        )

        # The original dropSpecs have changed as well
        a = graphDesc[0]
        c = graphDesc[2]

        self.assertEqual(1, len(a["consumers"]))
        self.assertEqual("B", a["consumers"][0])
        self.assertFalse("producers" in a and len(a["producers"]) > 0)
        self.assertFalse(
            "streamingConsumers" in c and len(c["streamingConsumers"]) > 0
        )

    def test_removeUnmetRelationships_named(self):
        with (files(test_graphs) / "HelloWorld_simplePG.graph").open() as f:
            graphDesc = json.load(f)
        unmetRelationships = graph_loader.removeUnmetRelationships(graphDesc)
        self.assertEqual(0, len(unmetRelationships))

    def test_unNamedPorts(self):
        """
        Use a graph with un-named ports and check whether it is loading
        """
        with (files(test_graphs) / "funcTestPG.graph").open() as f:
            graphSpec = json.load(f)
        a = graph_loader.createGraphFromDropSpecList(graphSpec)
        dummy = a

    def test_namedPorts(self):
        """
        Use a graph with named ports and check whether it is loading
        """
        with (files(test_graphs) / "funcTestPG_namedPorts.graph").open() as f:
            graphSpec = json.load(f)
        a = graph_loader.createGraphFromDropSpecList(graphSpec)
        dummy = a

    def test_applicationArgs(self):
        """
        Use a graph with applicationArgs and make sure applications see their
        arguments
        """
        with (files(test_graphs) / "application_args.graph").open() as f:
            graphSpec = json.load(f)
        graph = graph_loader.createGraphFromDropSpecList(graphSpec)
        app = graph[0]
        self.assertEqual(app.__class__, RandomArrayApp)
        self.assertEqual(app.size, 50)
        self.assertEqual(app.integer, True)
        self.assertEqual(app.low, 34)
        self.assertEqual(app.high, 3456)

    def test_backwardsCompatibilityWithPreciousFlag(self):
        """
        Ensure that precious is detected and exposed as the persist flag
        """
        testCases = [
            ("precious", True),
            ("precious", False),
            ("persist", True),
            ("persist", False),
            (None, False),  # Default of False is specific to MemoryDrops
        ]
        for key, value in testCases:
            with self.subTest(key=key, value=value):
                dropSpec = {
                    "oid": "A",
                    "categoryType": "Data",
                    "dropclass": "dlg.data.drops.memory.InMemoryDROP",
                }
                if key is not None:
                    dropSpec[key] = value

                graph = graph_loader.createGraphFromDropSpecList([dropSpec])
                data = graph[0]
                self.assertIsInstance(data, InMemoryDROP)
                self.assertEqual(value, data.persist)
                self.assertFalse(hasattr(data, "precious"))



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing the NodeManager, which directly manages DROP instances, and
thus represents the bottom of the DROP management hierarchy.
"""

from __future__ import annotations
import abc
import collections
import copy
import logging
from typing import Optional

from psutil import cpu_count
import os
import queue
import signal
import sys
import threading
import time
import typing
# import dlg.runtime.dlg_logging as logging


from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future

from dlg import constants
from dlg.manager.drop_manager import DROPManager
from dlg.manager.session import Session

from dlg import rpc, utils
from dlg.ddap_protocol import DROPStates
from dlg.apps.app_base import AppDROP, DropRunner
from dlg.exceptions import (
    NoSessionException,
    SessionAlreadyExistsException,
    DaliugeException, ErrorManagerCaughtException, SessionInterruptError,
)
from ..lifecycle.dlm import DataLifecycleManager


logger = logging.getLogger(f"dlg.{__name__}")

class NMDropEventListener(object):
    def __init__(self, nm, session_id):
        self._nm = nm
        self._session_id = session_id

    def handleEvent(self, event):
        event.session_id = self._session_id
        self._nm.publish_event(event)


class LogEvtListener(object):
    def handleEvent(self, event):
        if event.type == "status":
            logger.debug(
                "Data Drop uid=%s, oid=%s changed to state %s",
                event.uid,
                event.oid,
                event.status,
            )
        elif event.type == "execStatus":
            logger.debug(
                "AppDrop uid=%s, oid=%s changed to execState %s",
                event.uid,
                event.oid,
                event.execStatus,
            )


class ErrorStatusListener(object):
    """An event listener that passes down the erroneous drop to an error handler"""

    def __init__(self, session, error_listener):
        self._session = session
        self._error_listener = error_listener

    def handleEvent(self, evt):
        if evt.type == "status" and evt.status == DROPStates.ERROR:
            self._error_listener.on_error(self._session.drops[evt.uid])


def _load(obj, callable_attr):
    """
    Returns object (a python object or a string denoting a class within a
    python module only if it has the indicated attribute and it is callable.
    """
    if isinstance(obj, str):
        obj = utils.get_symbol(obj)()
    if not hasattr(obj, callable_attr) or not callable(getattr(obj, callable_attr)):
        raise ValueError(
            "%r doesn't contain an %s attribute that can be called"
            % (obj, callable_attr)
        )
    return obj


class NodeManagerDropRunner(DropRunner):
    @abc.abstractmethod
    def start(self, rpc_endpoint):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self):
        raise NotImplementedError


class NodeManagerThreadDropRunner(NodeManagerDropRunner):
    def __init__(self, max_workers: int):
        self._max_workers = max_workers
        self._thread_pool: Optional[ThreadPoolExecutor] = None

    def start(self, _rpc_endpoint):
        logger.info("Initializing thread pool with %d workers", self._max_workers)
        self._thread_pool = ThreadPoolExecutor(max_workers=self._max_workers)

    def run_drop(self, app_drop: AppDROP) -> Future:
        if logger.getEffectiveLevel() != logging.getLevelName(app_drop.log_level):
            logging.getLogger("dlg").setLevel(app_drop.log_level)
            logger.warning(
                "Set log-level for %s.%s to %s",
                app_drop.name,
                app_drop.humanKey,
                logging.getLevelName(logger.getEffectiveLevel()),
            )
        return self._thread_pool.submit(
            app_drop.run,
        )

    def close(self):
        self._thread_pool.shutdown(wait=True)


class NodeManagerProcessDropRunner(NodeManagerDropRunner):
    # Process isolated properties - should only be accessed in @classmethods
    # to ensure that they are global to a single process only
    _rpc_client: typing.Optional[rpc.RPCClient]
    _rpc_endpoint: typing.Tuple[str, int]

    def __init__(self, max_workers: int):
        self._max_workers = max_workers
        self._process_pool: typing.Optional[ProcessPoolExecutor] = None

    def start(self, rpc_endpoint):
        logger.info("Initializing process pool with %d workers", self._max_workers)

        self._process_pool = ProcessPoolExecutor(
            max_workers=self._max_workers,
            initializer=NodeManagerProcessDropRunner._setup_process,
            initargs=(rpc_endpoint,),
        )

    @classmethod
    def _setup_process(cls, rpc_endpoint):
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        cls._rpc_endpoint = rpc_endpoint
        cls._rpc_client = rpc.RPCClient()

        # We'll just let this be cleaned up when the process terminates
        # (instead of the normal `shutdown()`)
        cls._rpc_client.start()

    def run_drop(self, app_drop: AppDROP):
        inputs_proxy_info, outputs_proxy_info = (
            NodeManagerProcessDropRunner._get_proxy_infos(app_drop)
        )

        # MP Queues pickle on a background thread - we need to ensure that we don't
        # modify the input app_drop reference outside of the scope of this method
        # so we create a shallow copy that can be pickled at an indeterminate point in the future
        # See https://github.com/python/cpython/blob/v3.10.13/Lib/multiprocessing/queues.py#L95
        copied_drop = copy.copy(app_drop)
        copied_drop.inputs = collections.OrderedDict()
        copied_drop.outputs = collections.OrderedDict()

        return self._process_pool.submit(
            NodeManagerProcessDropRunner._run_app_drop,
            copied_drop,
            inputs_proxy_info,
            outputs_proxy_info,
        )

    @classmethod
    def _run_app_drop(cls, app_drop, inputs_proxy_info, outputs_proxy_info):
        cls._setup_drop_proxies(app_drop, inputs_proxy_info, outputs_proxy_info)
        return app_drop.run()

    @classmethod
    def _setup_drop_proxies(
        cls, app_drop: AppDROP, inputs_proxy_info, outputs_proxy_info
    ):
        app_drop.rpc_endpoint = cls._rpc_endpoint
        for input_proxy_info in inputs_proxy_info:
            app_drop.addInput(
                rpc.DropProxy(cls._rpc_client, input_proxy_info), back=False
            )
        for output_proxy_info in outputs_proxy_info:
            app_drop.addOutput(
                rpc.DropProxy(cls._rpc_client, output_proxy_info), back=False
            )

    @classmethod
    def _get_proxy_infos(cls, app_drop):
        inputs = [rpc.ProxyInfo.from_data_drop(i) for i in app_drop.inputs]
        outputs = [rpc.ProxyInfo.from_data_drop(o) for o in app_drop.outputs]
        return inputs, outputs

    def close(self):
        self._process_pool.shutdown(wait=True)
        logger.info("Process pool closed")


class NodeManagerBase(DROPManager):
    """
    Base class for a DROPManager that creates and holds references to DROPs.

    A NodeManagerBase is the ultimate responsible of handling DROPs. It does so not
    directly, but via Sessions, which represent and encapsulate separate,
    independent DROP graph executions. All DROPs created by the
    different Sessions are also given to a common DataLifecycleManager, which
    takes care of expiring them when needed and replicating them.

    Since a NodeManagerBase can handle more than one session, in principle only one
    NodeManagerBase is needed for each computing node, thus its name.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        events_port,
        dlm_check_period=0,
        dlm_cleanup_period=0,
        dlm_enable_replication=False,
        dlgPath=None,
        error_listener=None,
        event_listeners: list = None,
        max_threads=0,
        use_processes=False,
        logdir=utils.getDlgLogsDir(),
        use_local_time=False
    ):
        self._events_port = events_port
        self._dlm = DataLifecycleManager(
            check_period=dlm_check_period,
            cleanup_period=dlm_cleanup_period,
            enable_drop_replication=dlm_enable_replication,
        )
        self.sessions = {}
        self.logdir = logdir
        self.use_local_time=use_local_time

        # dlgPath may contain code added by the user with possible
        # DROP applications
        if dlgPath:
            dlgPath = os.path.expanduser(dlgPath)
            if os.path.isdir(dlgPath):
                logger.info("Adding %s to the system path", dlgPath)
                sys.path.append(dlgPath)
                # we also add underlying site-packages dir to support
                # the --prefix installation of code
                pyVer = f"{sys.version_info.major}.{sys.version_info.minor}"
                extraPath = f"{dlgPath}/lib/python{pyVer}/site-packages"
                logger.info("Adding %s to the system path", extraPath)
                sys.path.append(extraPath)

        # Error listener used by users to deal with errors coming from specific
        # Drops in whatever way they want. This is a specific case of an event
        # listener, so we add it together with the rest of the user-supplied
        # event listeners
        self._error_listener = (
            _load(error_listener, "on_error") if error_listener else None
        )
        if not event_listeners:
            event_listeners = []
        self._event_listeners = [_load(l, "handleEvent") for l in event_listeners]

        if max_threads <= 0:
            max_threads = cpu_count(logical=False)

        self.drop_runner: NodeManagerDropRunner
        if use_processes:
            self.drop_runner = NodeManagerProcessDropRunner(max_threads)
        else:
            self.drop_runner = NodeManagerThreadDropRunner(max_threads)

        # Event handler that only logs status changes
        debugging = logger.isEnabledFor(logging.DEBUG)
        self._logging_event_listener = LogEvtListener() if debugging else None

    def start(self, rpc_endpoint):
        super().start()
        logger.user("Running NodeManager")
        self.drop_runner.start(rpc_endpoint)
        self._dlm.startup()

    def shutdown(self):
        logger.user("Stopping NodeManager")
        self._dlm.cleanup()
        self.drop_runner.close()
        super().shutdown()

    def deliver_event(self, evt):
        """
        Method called by subclasses when a new event has arrived through the
        subscription mechanism.
        """
        if not evt.session_id in self.sessions:
            logger.warning(
                "No session %s found, event (%s) will be dropped",
                evt.session_id,
                evt.type,
            )
            return
        self.sessions[evt.session_id].deliver_event(evt)

    def _check_session_id(self, session_id):
        if session_id not in self.sessions:
            raise NoSessionException(session_id)

    def createSession(self, sessionId):
        if sessionId in self.sessions:
            raise SessionAlreadyExistsException(sessionId)
        self.sessions[sessionId] = Session(sessionId, nm=self)
        logger.info("Created session %s", sessionId)

    def getSessionStatus(self, sessionId):
        try:
            self._check_session_id(sessionId)
        except NoSessionException:
            return None

        return self.sessions[sessionId].status

    def getSessionReproStatus(self, sessionId):
        return self.sessions[sessionId].reprostatus

    def getSessionDir(self, sessionId):
        return {"dir" : self.sessions[sessionId].sessionDir}

    def getGraphReproData(self, sessionId):
        return self.sessions[sessionId].graphreprodata

    def linkGraphParts(self, sessionId, lhOID, rhOID, linkType):
        self._check_session_id(sessionId)
        self.sessions[sessionId].linkGraphParts(lhOID, rhOID, linkType)

    def addGraphSpec(self, sessionId, graphSpec):
        self._check_session_id(sessionId)
        self.sessions[sessionId].addGraphSpec(graphSpec)

    def getGraphStatus(self, sessionId):
        self._check_session_id(sessionId)
        return self.sessions[sessionId].getGraphStatus()

    def getDropStatus(self, sessionId, dropId):
        return self.sessions[sessionId].getDropLogs(dropId)

    def getGraph(self, sessionId):
        self._check_session_id(sessionId)
        #  TODO: Ensure returns reproducibility data.
        return self.sessions[sessionId].getGraph()

    def getLogDir(self):
        return self.logdir

    from dlg.runtime.error_management import manage_session_failure

    @manage_session_failure
    def deploySession(self, sessionId, completedDrops: list[str] = None):
        self._check_session_id(sessionId)
        session = self.sessions[sessionId]
        if hasattr(self, "_memoryManager"):
            self._memoryManager.register_session(sessionId)

        def foreach(drop):
            drop.autofill_environment_variables()
            drop.drop_runner = self.drop_runner
            self._dlm.addDrop(drop)

            # Remote event forwarding
            evt_listener = NMDropEventListener(self, sessionId)
            if isinstance(drop, AppDROP):
                drop.subscribe(evt_listener, "producerFinished")
            else:
                drop.subscribe(evt_listener, "dropCompleted")

            # Purely for logging purposes
            log_evt_listener = self._logging_event_listener
            if log_evt_listener:
                drop.subscribe(log_evt_listener, "status")
                drop.subscribe(log_evt_listener, "reproducibility")
                if isinstance(drop, AppDROP):
                    drop.subscribe(log_evt_listener, "execStatus")

        # Add user-supplied listeners
        listeners = self._event_listeners[:]
        if self._error_listener:
            listeners.append(ErrorStatusListener(session, self._error_listener))

        try:
            session.deploy(
                completedDrops=completedDrops,
                event_listeners=listeners,
                foreach=foreach,
            )
        except ErrorManagerCaughtException as exc:
            session.cancel(interrupt=True)
            raise SessionInterruptError(
                "Session was interrupted due to Deploy failure"
            ) from exc

    def cancelSession(self, sessionId):
        logger.info("Cancelling session: %s", sessionId)
        self._check_session_id(sessionId)
        self.sessions[sessionId].cancel()

    def destroySession(self, sessionId):
        logger.info("Destroying session: %s", sessionId)
        self._check_session_id(sessionId)
        session = self.sessions.pop(sessionId)
        if hasattr(self, "_memoryManager"):
            self._memoryManager.shutdown_session(sessionId)
        self._dlm.remove_drops(session.drops)
        session.destroy()

    def getSessionIds(self):
        return list(self.sessions.keys())

    def getGraphSize(self, sessionId):
        try:
            self._check_session_id(sessionId)
        except NoSessionException:
            return 0
        session = self.sessions[sessionId]
        return len(session.graph)

    def trigger_drops(self, sessionId, uids):
        self._check_session_id(sessionId)
        t = threading.Thread(
            target=self.sessions[sessionId].trigger_drops,
            name="Drop trigger",
            args=(uids,),
        )
        t.start()

    def add_node_subscriptions(self, sessionId, relationships):
        logger.debug("Received subscription information: %r", relationships)
        self._check_session_id(sessionId)
        self.sessions[sessionId].add_node_subscriptions(relationships)

        # Set up event channels subscriptions
        for nodesub in relationships:
            # node = Node(nodesub)
            # This needs to be changed
            events_port = nodesub.events_port  # constants.NODE_DEFAULT_EVENTS_PORT
            if type(nodesub) is tuple:
                host, events_port, _ = nodesub
            else:
                # TODO: we also have to unsubscribe from them at some point
                # if nodesub.find(":") > 0:
                #     host, _ = nodesub.split(":")
                # else:
                host = nodesub
            logger.debug("Sending subscription to %s", f"{host}:{events_port}")
            self.subscribe(host, events_port)

    def has_method(self, sessionId, uid, mname):
        self._check_session_id(sessionId)
        return self.sessions[sessionId].has_method(uid, mname)

    def get_drop_property(self, sessionId, uuid, prop_name):
        self._check_session_id(sessionId)
        return self.sessions[sessionId].get_drop_property(uuid, prop_name)

    def call_drop(self, sessionId, uid, method, *args):
        self._check_session_id(sessionId)
        return self.sessions[sessionId].call_drop(uid, method, *args)


class ZMQPubSubMixIn(object):
    """
    ZeroMQ-based event publisher and subscriber.

    Event publishing and event reception are done in their own separate
    threads, where the externally-facing ZeroMQ sockets are created and used.

    Events to be published are fed into the publishing thread via a safe-thread
    Queue object (self._events_out), enabling any local thread to publish events
    without having to worry about ZeroMQ thread-safeness.

    The event reception thread not only *receives* events, but also updates the
    subscription socket to connect to new peers. These updates are fed via a
    Queue object (self._subscriptions), enabling any local thread to indicate a
    new peer to subscribe to in a thread-safe manner.

    Note that we investigated not using Queue objects to communicate between
    threads, and use inproc:// ZeroMQ sockets instead. This works, but at a
    cost: all threads putting values into these sockets would need to check,
    each time they use a socket in any manner, if the Context object is still
    valid and hasn't been closed (or alternatively if self._pubsub_running is
    still True). Our experience with this alternative was not satisfactory, and
    therefore we went for a Queue-based thread communication model, making the
    handling of ZeroMQ resources simpler.
    """

    subscription = collections.namedtuple("subscription", "endpoint finished_evt")

    def __init__(self, host, events_port):
        self._events_host = host
        self._events_port = events_port

    def start(self):
        self._pubsub_running = True
        super(ZMQPubSubMixIn, self).start()
        self._events_in = queue.Queue()
        self._events_out = queue.Queue()
        self._subscriptions = queue.Queue()

        # Starts background threads, but wait until their sockets are created
        timeout = 10
        self._event_publisher = self._start_thread(
            self._publish_events, "Evt pub", timeout
        )
        self._event_receiver = self._start_thread(
            self._receive_events, "Evt recv", timeout
        )
        self._event_deliverer = self._start_thread(self._deliver_events, "Evt delivery")

    def _start_thread(self, target, name, timeout=None):
        evt = threading.Event() if timeout else None
        args = (evt,) if evt else ()
        t = threading.Thread(target=target, name=name, args=args)
        t.start()
        if evt and not evt.wait(timeout):
            raise Exception("Failed to start %s thread in %d seconds" % (name, timeout))
        return t

    def shutdown(self):
        super(ZMQPubSubMixIn, self).shutdown()
        self._pubsub_running = False
        self._event_deliverer.join()
        self._event_publisher.join()
        self._event_receiver.join()
        logger.info("ZeroMQ event publisher/subscriber finished")

    def publish_event(self, evt):
        self._events_out.put(evt)

    def subscribe(self, host, port):
        timeout = 5
        finished_evt = threading.Event()
        endpoint = "tcp://%s:%d" % (utils.zmq_safe(host.host), port)
        self._subscriptions.put(ZMQPubSubMixIn.subscription(endpoint, finished_evt))
        if not finished_evt.wait(timeout):
            raise DaliugeException(
                "ZMQ subscription not achieved within %d seconds" % (timeout,)
            )
        logger.info("Subscribed for events originating from %s", endpoint)

    def _publish_events(self, sock_created):
        import zmq

        pub = self._context.socket(zmq.PUB)  # @UndefinedVariable
        pub.set_hwm(0)  # Never drop messages that should be sent
        endpoint = "tcp://%s:%d" % (
            utils.zmq_safe(self._events_host),
            self._events_port,
        )
        pub.bind(endpoint)
        logger.info("Publishing events via ZeroMQ on %s", endpoint)
        sock_created.set()

        while self._pubsub_running:
            try:
                obj = self._events_out.get_nowait()
            except queue.Empty:
                time.sleep(0.01)
                continue

            while self._pubsub_running:
                try:
                    pub.send_pyobj(obj, flags=zmq.NOBLOCK)  # @UndefinedVariable
                    break
                except zmq.error.Again:
                    logger.debug("Got an 'Again' when publishing event")
                    time.sleep(0.01)
                    continue
        pub.close()

    def _deliver_events(self):
        while self._pubsub_running:
            try:
                evt = self._events_in.get_nowait()
                self.deliver_event(evt)
            except queue.Empty:
                time.sleep(0.01)

    def _receive_events(self, sock_created):
        import zmq
        from zmq.utils.monitor import recv_monitor_message

        sub = self._context.socket(zmq.SUB)  # @UndefinedVariable
        sub_endpoints = set()
        sub.setsockopt(zmq.SUBSCRIBE, b"")  # @UndefinedVariable
        sub_monitor = sub.get_monitor_socket()
        sock_created.set()

        pending_connections = {}
        while self._pubsub_running:
            # A new subscription has been requested
            try:
                subscription = self._subscriptions.get_nowait()
                if subscription.endpoint in sub_endpoints:
                    subscription.finished_evt.set()
                else:
                    sub.connect(subscription.endpoint)
                    pending_connections[subscription.endpoint] = (
                        subscription.finished_evt
                    )
            except queue.Empty:
                pass

            try:
                msg = recv_monitor_message(sub_monitor, flags=zmq.NOBLOCK)
                if msg["event"] != zmq.EVENT_CONNECTED:
                    continue
                endpoint = utils.b2s(msg["endpoint"])
                sub_endpoints.add(endpoint)
                finished_evt = pending_connections.pop(endpoint)
                finished_evt.set()
            except zmq.error.Again:
                pass

            try:
                evt = sub.recv_pyobj(flags=zmq.NOBLOCK)  # @UndefinedVariable
                self._events_in.put(evt)
            except zmq.error.Again:
                time.sleep(0.01)
            except zmq.error.ZMQError:
                # Figure out what to do here
                logger.exception(
                    "Something bad happened in %s:%d to ZMQ :'(",
                    self._events_host,
                    self._events_port,
                )
                break

        # Flush pending connection events to avoid callers hanging out forever
        for evt in pending_connections:
            evt.set()

        sub_monitor.close()
        sub.close()


# So far we currently support ZMQ only for event publishing
EventMixIn = ZMQPubSubMixIn


# Load the corresponding RPC classes and finish the construction of NodeManager
class RpcMixIn(rpc.RPCClient, rpc.RPCServer):
    pass


# Final NodeManager class
class NodeManager(NodeManagerBase, EventMixIn, RpcMixIn):
    def __init__( # pylint: disable=keyword-arg-before-vararg
        self,
        host=None,
        rpc_port=constants.NODE_DEFAULT_RPC_PORT,
        events_port=constants.NODE_DEFAULT_EVENTS_PORT,
        *args,
        **kwargs,
    ):
        host = host or "localhost"
        NodeManagerBase.__init__(self, events_port, *args, **kwargs)
        EventMixIn.__init__(self, host, events_port)
        RpcMixIn.__init__(self, host, rpc_port)
        self.start()

    def start(self): # pylint: disable=arguments-differ
        # We "just know" that our RpcMixIn will have a create_context static
        # method, which in reality means we are using the ZeroRPCServer class
        self._context = RpcMixIn.create_context()
        super().start(self.rpc_endpoint)

    def shutdown(self):
        super(NodeManager, self).shutdown()
        self._context.term()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Replicated from utils.py in order to be able to call it stand-alone from
run_engine.sh. This function fills template user and group files, which will
then be used inside the docker containers to ensure that the user running the
app inside the container is the same as outside.
"""

import os
import pwd
import grp
import platform


def prepareUser(DLG_ROOT="."):
    workdir = f"{DLG_ROOT}/workspace/settings"
    try:
        os.makedirs(workdir, exist_ok=True)
    except OSError as e:
        raise e
    template_dir = os.path.dirname(__file__)
    # get current user info
    pw = pwd.getpwuid(os.getuid())
    gr = grp.getgrgid(pw.pw_gid)
    if platform.system() == "Darwin":
        grpnam = "staff"
    else:
        grpnam = "docker"
    dgr = grp.getgrnam(grpnam)
    with open(os.path.join(workdir, "passwd"), "wt") as file:
        file.write(open(os.path.join(template_dir, "passwd.template"), "rt").read())
        file.write(
            f"{pw.pw_name}:x:{pw.pw_uid}:{pw.pw_gid}:{pw.pw_gecos}:{DLG_ROOT}:/bin/bash\n"
        )
    with open(os.path.join(workdir, "group"), "wt") as file:
        file.write(open(os.path.join(template_dir, "group.template"), "rt").read())
        file.write(f"{gr.gr_name}:x:{gr.gr_gid}:\n")
        file.write(f"docker:x:{dgr.gr_gid}:{pw.pw_name}\n")
        file.write(f"sudo:x:27:{gr.gr_name}\n")

    return dgr.gr_gid
